/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include <BackupDialog.h>

// Qt includes
#include <QDebug>
#include <QFileInfo>
#include <QRegularExpression>
#include <QMessageBox>
#include <QDir>
#include <QVBoxLayout>
#include <QListWidget>
#include <QLabel>

namespace {
const QString BACKUP_DATE_FORMAT = QStringLiteral("yyyy-MM-dd_hh-mm-ss");
const QString DATE_DATE_FORMAT   = QStringLiteral("dd.MM.yyyy");
const QString TIME_DATE_FORMAT   = QStringLiteral("hh:mm:ss");
const QString FULL_DATE_FORMAT   = QStringLiteral("dd.MM.yyyy hh:mm:ss");
}

BackupDialog::BackupDialog(QWidget *parent, const QString &dbFileName)
    : Dialog(parent, nullptr, Dialog::DeleteMode::NoDelete)
{
    setTitle(tr("Backup wiederherstellen"));

    m_dbToOverwrite = dbFileName;

    QLabel *header = new QLabel;
    header->setWordWrap(true);
    layout()->addWidget(header);

    m_backups = new QListWidget;
    layout()->addWidget(m_backups);
    connect(m_backups, &QListWidget::itemSelectionChanged, this, &BackupDialog::enableOkButton);

    setButtonText(QDialogButtonBox::Ok, tr("Wiederherstellen"));
    setOkButtonEnabled(false);

    // Search for Backups and fill in the data

    BackupDialog::DbFileInfo dbFileInfo = parseDbFileName(dbFileName);
    QString baseName = dbFileInfo.baseName;
    if (! dbFileInfo.originalBaseName.isEmpty()) {
        if (QMessageBox::question(this,
                tr("Backup wiederherstellen"),
                tr("<p><b>Ist ein Backup offen?</b></p>"
                   "<p>Es sieht so aus, als wäre die offene Turnierdatenbank „%1“ ein Backup von "
                   "„%2.mtdb“, das %3 erstellt wurde.</p>"
                   "<p>Soll nach Backups von „%2.mtdb“ gesucht werden?</p>").arg(
                   dbFileName, dbFileInfo.originalBaseName, getDateString(dbFileInfo.backupDate)),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes) {
            baseName = dbFileInfo.originalBaseName;
            m_dbToOverwrite = dbFileInfo.path + QStringLiteral("/")
                                                  + baseName + QStringLiteral(".mtdb");
        }
    }

    // We set the header text here because possibly, we don't search for
    // the open database if it's a backup but for the base db
    header->setText(tr("Für die Turnierdatenbank „%1“ wurden folgende Backups gefunden:").arg(
                       QDir::toNativeSeparators(m_dbToOverwrite)));

    QDir dir(dbFileInfo.path);
    QStringList filters({ baseName + QStringLiteral("_*.mtdb") });
    dir.setNameFilters(filters);
    QRegularExpression re(QStringLiteral("^")
                          + QRegularExpression::escape(baseName)
                          + QStringLiteral("_([0-9]{4}-[0-9]{2}-[0-9]{2}_"
                                           "[0-9]{2}-[0-9]{2}-[0-9]{2})_?([0-9]+)?\\.mtdb$"));

    for (const QString &file : dir.entryList(QDir::Files, QDir::Name)) {
        QRegularExpressionMatch match = re.match(file);
        if (match.hasMatch()) {
            QString date = getDateString(match.captured(1));
            if (! match.captured(2).isEmpty()) {
                date += QStringLiteral(" (%1)").arg(match.captured(2));
            }
            QListWidgetItem *item = new QListWidgetItem(date);
            item->setData(Qt::UserRole, QVariant(dbFileInfo.path + QStringLiteral("/") + file));
            m_backups->addItem(item);
        }
    }

    // Select the opened backup in case we have one and searched for the base db
    if (dbFileInfo.baseName != baseName) {
        for (int i = 0; i < m_backups->count(); i++) {
            if (m_backups->item(i)->data(Qt::UserRole).toString() == dbFileName) {
                m_backups->item(i)->setText(tr("%1 (jetzt geöffnet)").arg(
                                            m_backups->item(i)->text()));
                m_backups->setCurrentItem(m_backups->item(i));
                break;
            }
        }
    }

    if (m_backups->count() == 0) {
        m_backups->addItem(tr("Keine Backups gefunden"));
        m_backups->setEnabled(false);
    }
}

void BackupDialog::enableOkButton()
{
    if (! okButtonEnabled()) {
        setOkButtonEnabled(true);
    }
}

QString BackupDialog::backupToRestore() const
{
    return m_backups->currentItem()->data(Qt::UserRole).toString();
}

QString BackupDialog::dbToOverwrite() const
{
    return m_dbToOverwrite;
}

QString BackupDialog::createBackup(const QString &source, const QString &target,
                                   const QString &addition)
{
    QString base;
    if (target.isEmpty()) {
        DbFileInfo fileInfo = parseDbFileName(source);
        base = fileInfo.path + QStringLiteral("/") + fileInfo.baseName
               + QStringLiteral("_") + addition;
    } else {
        base = target + QStringLiteral("_") + addition;
    }
    QString backup = base + QStringLiteral(".mtdb");

    // Be sure to create a non-existant file
    int number = 0;
    while (QFileInfo::exists(backup)) {
        number++;
        backup = base + QStringLiteral("_%1.mtdb").arg(QString::number(number));
    }

    // Check if we can actually write the file. Yes, we have to do it in this way, because
    // QFileInfo::isWritable does not work as expected on Windows (returns false for
    // "My Documents").
    QFile test(backup);
    if (test.open(QIODevice::WriteOnly)) {
        test.close();
        test.remove(); // This is save because we can be sure that backup does not exist yet
    } else {
        return QString();
    }

    if (QFile::copy(source, backup)) {
        return backup;
    }

    return QString();
}

BackupDialog::DbFileInfo BackupDialog::parseDbFileName(const QString &dbFileName)
{
    QFileInfo fileInfo(dbFileName);
    DbFileInfo dbFileInfo;
    dbFileInfo.path = fileInfo.path();
    dbFileInfo.baseName = fileInfo.completeBaseName();
    QRegularExpression re(QStringLiteral("^(.+)_([0-9]{4}-[0-9]{2}-[0-9]{2}_"
                                         "[0-9]{2}-[0-9]{2}-[0-9]{2})(_[0-9]+)?$"));
    QRegularExpressionMatch match = re.match(dbFileInfo.baseName);
    if (match.hasMatch()) {
        dbFileInfo.originalBaseName = match.captured(1);
        dbFileInfo.backupDate = QDateTime::fromString(match.captured(2), BACKUP_DATE_FORMAT);
    }
    return dbFileInfo;
}

QString BackupDialog::getDateString(const QDateTime &backupDate)
{
    const QString currentDateString = QDate::currentDate().toString(DATE_DATE_FORMAT);
    const QString backupDateString = backupDate.toString(DATE_DATE_FORMAT);
    if (currentDateString == backupDateString) {
        return tr("um %1 Uhr").arg(backupDate.toString(TIME_DATE_FORMAT));
    } else {
        return tr("am %1 um %2 Uhr").arg(backupDateString, backupDate.toString(TIME_DATE_FORMAT));
    }
}

QString BackupDialog::getDateString(const QString &backupDate) const
{
    const QDateTime backupDateTime = QDateTime::fromString(backupDate, BACKUP_DATE_FORMAT);
    const QString currentDateString = QDate::currentDate().toString(DATE_DATE_FORMAT);
    const QString backupDateString = backupDateTime.toString(DATE_DATE_FORMAT);
    if (currentDateString == backupDateString) {
        return backupDateTime.toString(TIME_DATE_FORMAT);
    } else {
        return backupDateTime.toString(FULL_DATE_FORMAT);
    }
}

QString BackupDialog::createDatabaseBackup(const QString &source, const QString &target)
{
    return createBackup(source, target, QDateTime::currentDateTime().toString(BACKUP_DATE_FORMAT));
}
