/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "BoogerWidget.h"
#include "BoogerSpinBox.h"
#include "BoogerRadioButton.h"
#include "SharedStyles.h"

// Qt includes
#include <QDebug>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QCheckBox>
#include <QApplication>
#include <QStyle>

BoogerWidget::BoogerWidget(QWidget *parent, int booger, int boogersPerRound, int boogerScore)
    : QWidget(parent), m_boogerScore(boogerScore), m_boogerNumber(booger)
{
    setAutoFillBackground(false);

    QGridLayout *layout = new QGridLayout(this);

    QHBoxLayout *headerLayout = new QHBoxLayout;
    layout->addLayout(headerLayout, 0, 0, 1, 2);

    QLabel *boogerHeader = new QLabel(tr("%1. Bobbl").arg(QString::number(booger)));
    boogerHeader->setStyleSheet(QStringLiteral(STYLE_BOLD_TEXT));
    headerLayout->addWidget(boogerHeader);

    headerLayout->addStretch();

    m_boogerAborted = new QPushButton;
    m_boogerAborted->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCancelButton));
    m_boogerAborted->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_boogerAborted->setCheckable(true);
    m_boogerAborted->setFlat(true);
    m_boogerAborted->setToolTip(tr("Als abgebrochen markieren"));
    m_boogerAborted->setWhatsThis(tr(
        "<p>Wenn diese Option gesetzt ist, wird der Bobbl und alle folgenden (sofern vorhanden) "
        "als abgebrochen markiert (z.&thinsp;B. wenn mit einer Zeitbegrenzung pro Runde gespielt "
        "wird, und die Zeit abgelaufen ist).</p>"
        "<p>In einem abgebrochenen Bobbl können zwei Spielstände eingegeben werden, entsprechend "
        "dem jeweils erreichten Punktestand. Auswertungstechnisch haben beide Paare den Bobbl "
        "verloren und bekommen die Punkte als „geschossene Tore“ gutgeschrieben."));
    connect(m_boogerAborted, &QPushButton::clicked, this, &BoogerWidget::abortedButtonClicked);
    headerLayout->addWidget(m_boogerAborted);

    for (int i = 0; i < 2; i++) {
        layout->addWidget(new QLabel(tr("Paar %1").arg(i + 1)), i + 1, 0);

        BoogerSpinBox *spinBox = new BoogerSpinBox;
        spinBox->setMaximum(m_boogerScore - (i == 0));
        connect(spinBox, &BoogerSpinBox::scoreDialogRequested,
                this, &BoogerWidget::scoreDialogRequested);
        if (booger == boogersPerRound) {
            connect(spinBox, &BoogerSpinBox::returnPressed,
                    this, &BoogerWidget::returnPressed);
        }

        BoogerRadioButton *radioButton = new BoogerRadioButton(this, spinBox);
        connect(radioButton, &QRadioButton::clicked, this, &BoogerWidget::toggleScore);
        connect(radioButton, &BoogerRadioButton::scoreDialogRequested,
                this, &BoogerWidget::scoreDialogRequested);

        m_score.append(spinBox);
        m_booger.append(radioButton);
        layout->addWidget(spinBox, i + 1, 1);
        layout->addWidget(radioButton, i + 1, 2);
    }

    m_booger.at(0)->setChecked(true);
    m_score.at(1)->setValue(m_boogerScore);
    m_score.at(1)->setEnabled(false);
}

void BoogerWidget::toggleScore()
{
    for (int i = 0; i < 2; i++) {
        if (m_booger.at(i)->isChecked()) {
            m_score.at(i)->setMaximum(m_boogerScore - 1);
            m_score.at(i)->setValue(0);
            m_score.at(i)->setEnabled(true);
            m_score.at(i)->setFocus();
            m_score.at(i)->selectAll();
        } else {
            m_score.at(i)->setMaximum(m_boogerScore);
            m_score.at(i)->setValue(m_boogerScore);
            m_score.at(i)->setEnabled(false);
        }
    }
}

int BoogerWidget::score(int pair) const
{
    return m_score.at(pair - 1)->value();
}

void BoogerWidget::setScore(int pair1Score, int pair2Score)
{
    if (pair1Score == m_boogerScore || pair2Score == m_boogerScore) {
        m_booger.at(0)->setChecked(pair1Score != m_boogerScore);
        m_booger.at(1)->setChecked(pair2Score != m_boogerScore);
        toggleScore();
    }
    m_score.at(0)->setValue(pair1Score);
    m_score.at(1)->setValue(pair2Score);
}

void BoogerWidget::abortedButtonClicked(bool aborted)
{
    emit abortedStateChanged(m_boogerNumber - 1, aborted);
    setAborted(aborted);
    if (aborted) {
        setAsFirstAborted();
        m_score.at(0)->setFocus();
        m_score.at(0)->selectAll();
    }
}

void BoogerWidget::setAsFirstAborted()
{
    m_score.at(0)->setEnabled(true);
    m_score.at(1)->setEnabled(true);
}

void BoogerWidget::setAborted(bool aborted)
{
    if (m_boogerAborted->isChecked() != aborted) {
        m_boogerAborted->setChecked(aborted);
    }
    if (aborted) {
        for (int i = 0; i < 2; i++) {
            m_booger.at(i)->setAutoExclusive(false);
            m_booger.at(i)->setChecked(true);
            m_booger.at(i)->setEnabled(false);
            m_score.at(i)->setEnabled(true);
            m_score.at(i)->setValue(0);
            m_score.at(i)->setMaximum(m_boogerScore - 1);
            m_score.at(i)->setEnabled(false);
        }
    } else {
        for (int i = 0; i < 2; i++) {
            m_booger.at(i)->setAutoExclusive(true);
            m_booger.at(i)->setEnabled(true);
        }
        m_booger.at(1)->setChecked(false);
        toggleScore();
    }
}

void BoogerWidget::reset()
{
    if (m_boogerAborted->isChecked()) {
        setAborted(false);
    }
    m_booger.at(0)->setChecked(true);
    toggleScore();
}
