/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ChecksumsDialog.h"
#include "RankingPage.h"
#include "TableDelegate.h"
#include "Cell.h"

// Qt includes
#include <QCryptographicHash>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QLabel>
#include <QGridLayout>
#include <QDebug>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>

ChecksumsDialog::ChecksumsDialog(QWidget *parent, Database *database)
    : Dialog(parent, database)
{
    setTitle(tr("Prüfsummen"));
    showOnlyOkButton();

    QGroupBox *generalChecksumsBox = new QGroupBox;
    generalChecksumsBox->setWhatsThis(tr(
        "<p>Sollen an mehreren Rechnern eingegebene Ergebnisse abgeglichen werden, um "
        "Eingabefehler zu finden, dann können Prüfsummen den Abgleich einfacher machen.</p>"
        "<p>Zunächst werden die Prüfsummen der %1listen verglichen. Diese müssen übereinstimmen, "
        "da die Namen mit in die Berechnung der Ranglisten-Prüfsumme(n) einfließen.</p>"
        "<p>Dann werden die Prüfsummen der Ranglisten verglichen. Sind die Prüfsummen "
        "identisch, dann wurden auch alle Ergebnisse gleich eingegeben. Weichen sie ab, dann "
        "kann die Rangliste weiter unterteilt werden, um den oder die Fehler zu finden "
        "(s.&thinsp;u.).</p>").arg(db()->isPairMode() ? tr("Paar") : tr("Spieler")));
    layout()->addWidget(generalChecksumsBox);
    QGridLayout *generalChecksumsLayout = new QGridLayout(generalChecksumsBox);

    // Pairs/Players list checksum

    QLabel *playersListLabel = new QLabel(QStringLiteral("<b>%1</b>").arg(
                                                         db()->isPairMode() ? tr("Paarliste:")
                                                                            : tr("Spielerliste:")));
    playersListLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    generalChecksumsLayout->addWidget(playersListLabel, 0, 0);

    QStringList playersList = db()->listPlayers();
    playersList.sort();
    QLabel *playersChecksum = new QLabel(checksum(playersList.join(QString())));
    playersChecksum->setTextInteractionFlags(Qt::TextSelectableByMouse);
    generalChecksumsLayout->addWidget(playersChecksum, 0, 1);

    // Whole ranking checksum

    QLabel *rankingLabel = new QLabel(QStringLiteral("<b>%1</b>").arg(tr("Rangliste:")));
    rankingLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    generalChecksumsLayout->addWidget(rankingLabel, 1, 0);

    QLabel *wholeRanking = new QLabel;
    generalChecksumsLayout->addWidget(wholeRanking, 1, 1);

    if (db()->getScores(1).count() == 0) {
        wholeRanking->setText(QStringLiteral("<i>%1</i>").arg(
                                                 tr("Bisher wurden keine Ergebnisse eingetragen")));
        return;
    } else {
        wholeRanking->setTextInteractionFlags(Qt::TextSelectableByMouse);
        m_ranking = db()->getRanking(db()->roundsState().count());
        m_firstStartRow = 0;
        m_firstRowCount = m_ranking.count();
        wholeRanking->setText(calculateChecksum(m_firstStartRow, m_firstRowCount));
    }

    // The partial ranking checksum GUI is only set up if we actually have a ranking

    m_showPartialChecksums = new QPushButton(tr("Teil-Prüfsummen der Rangliste anzeigen"));
    m_showPartialChecksums->setWhatsThis(tr(
        "<p>Wenn die Gesamt-Prüfsummen der Ranglisten nicht übereinstimmen, dann können Teil-"
        "Prüfsummen für die jeweilige Hälfte der Rangliste berechnet und abgeglichen werden.</p>"
        "<p>Stimmt die Prüfsumme für die erste Hälte nicht überein, dann wird diese weiter "
        "unterteilt. Stimmt die erste, aber nicht die zweite, dann ist in der ersten Hälfte alles "
        "okay, und die zweite Hälte wird weiter unterteilt, um die Falscheingabe zu finden.</p>"
    ));
    connect(m_showPartialChecksums, &QPushButton::clicked,
            this, &ChecksumsDialog::showPartialChecksums);
    layout()->addWidget(m_showPartialChecksums);

    m_reset = new QPushButton(tr("Teil-Prüfsummen\nzurücksetzen"));
    m_reset->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    generalChecksumsLayout->addWidget(m_reset, 0, 2, 2, 1, Qt::AlignRight);
    m_reset->hide();
    connect(m_reset, &QPushButton::clicked, this, &ChecksumsDialog::resetPartialChecksums);

    m_partialChecksumsBox = new QGroupBox(tr("Teil-Prüfsummen der Rangliste"));
    layout()->addWidget(m_partialChecksumsBox);
    m_partialChecksumsBox->hide();
    QVBoxLayout *partialChecksumsLayout = new QVBoxLayout(m_partialChecksumsBox);

    // First part of the ranking

    m_firstPartBox = new QGroupBox;
    partialChecksumsLayout->addWidget(m_firstPartBox);
    QVBoxLayout *fistPartLayout = new QVBoxLayout(m_firstPartBox);
    QHBoxLayout *firstPartChecksumLayout = new QHBoxLayout;
    fistPartLayout->addLayout(firstPartChecksumLayout);
    firstPartChecksumLayout->addWidget(new QLabel(QStringLiteral("<b>%1</b>").arg(
                                                                 tr("Prüfsumme:"))));
    m_firstPartChecksum = new QLabel;
    m_firstPartChecksum->setTextInteractionFlags(Qt::TextSelectableByMouse);
    firstPartChecksumLayout->addWidget(m_firstPartChecksum);
    firstPartChecksumLayout->addStretch();
    m_divideFirstPart = new QPushButton(tr("Aufteilen"));
    firstPartChecksumLayout->addWidget(m_divideFirstPart);
    connect(m_divideFirstPart, &QPushButton::clicked, this, &ChecksumsDialog::divideFirstPart);
    m_firstPartRanking = createTableWidget();
    fistPartLayout->addWidget(m_firstPartRanking);

    // Second part of the ranking

    m_secondPartBox = new QGroupBox;
    partialChecksumsLayout->addWidget(m_secondPartBox);
    QVBoxLayout *secondPartLayout = new QVBoxLayout(m_secondPartBox);
    QHBoxLayout *secondPartChecksumLayout = new QHBoxLayout;
    secondPartLayout->addLayout(secondPartChecksumLayout);
    secondPartChecksumLayout->addWidget(new QLabel(QStringLiteral("<b>%1</b>").arg(
                                                                  tr("Prüfsumme:"))));
    m_secondPartChecksum = new QLabel;
    m_secondPartChecksum->setTextInteractionFlags(Qt::TextSelectableByMouse);
    secondPartChecksumLayout->addWidget(m_secondPartChecksum);
    secondPartChecksumLayout->addStretch();
    m_divideSecondPart = new QPushButton(tr("Aufteilen"));
    secondPartChecksumLayout->addWidget(m_divideSecondPart);
    connect(m_divideSecondPart, &QPushButton::clicked, this, &ChecksumsDialog::divideSecondPart);
    m_secondPartRanking = createTableWidget();
    secondPartLayout->addWidget(m_secondPartRanking);
}

void ChecksumsDialog::showPartialChecksums()
{
    m_showPartialChecksums->hide();
    m_partialChecksumsBox->show();
    divideFirstPart();
    m_reset->setEnabled(false);
    m_reset->show();
}

const QString ChecksumsDialog::checksum(const QString &text) const
{
    QString checksum = QString::fromUtf8(QCryptographicHash::hash(text.toUtf8(),
                                         QCryptographicHash::Md5).toHex());
    for (int i = 4; i < checksum.size(); i += 5) {
        checksum.insert(i, QStringLiteral(" "));
    }
    return QStringLiteral("<pre>%1</pre>").arg(checksum);
}

const QString ChecksumsDialog::calculateChecksum(int startRow, int rowCount) const
{
    QString rankingData;
    for (int i = startRow; i < startRow + rowCount; i++) {
        rankingData += m_ranking.at(i).name
                    + QString::number(m_ranking.at(i).boogers)
                    + QString::number(m_ranking.at(i).goals)
                    + QString::number(m_ranking.at(i).opponentGoals);
    }
    return checksum(rankingData);
}

void ChecksumsDialog::updateDividedChecksums()
{
    if (m_firstRowCount > 1) {
        m_firstPartBox->setTitle(tr("Zeilen %1–%2").arg(
                                    QString::number(m_firstStartRow + 1),
                                    QString::number(m_firstStartRow + m_firstRowCount)));
    } else {
        m_firstPartBox->setTitle(tr("Zeile %1").arg(m_firstStartRow + 1));
        m_divideFirstPart->setEnabled(false);
    }

    if (m_secondRowCount > 1) {
        m_secondPartBox->setTitle(tr("Zeilen %1–%2").arg(
                                     QString::number(m_secondStartRow + 1),
                                     QString::number(m_secondStartRow + m_secondRowCount)));
    } else {
        m_secondPartBox->setTitle(tr("Zeile %1").arg(m_secondStartRow + 1));
        m_divideSecondPart->setEnabled(false);
    }

    m_firstPartChecksum->setText(calculateChecksum(m_firstStartRow, m_firstRowCount));
    m_secondPartChecksum->setText(calculateChecksum(m_secondStartRow, m_secondRowCount));

    updateRankingTable(m_firstPartRanking, m_firstStartRow, m_firstRowCount);
    updateRankingTable(m_secondPartRanking, m_secondStartRow, m_secondRowCount);
}

void ChecksumsDialog::divideFirstPart()
{
    m_secondRowCount = m_firstRowCount / 2;
    m_firstRowCount = m_firstRowCount / 2 + m_firstRowCount % 2;
    m_secondStartRow = m_firstStartRow + m_firstRowCount;
    updateDividedChecksums();
    m_reset->setEnabled(true);
}

void ChecksumsDialog::divideSecondPart()
{
    m_firstStartRow = m_secondStartRow;
    m_firstRowCount = m_secondRowCount / 2 + m_secondRowCount % 2;
    m_secondRowCount = m_secondRowCount / 2;
    m_secondStartRow = m_firstStartRow + m_firstRowCount;
    updateDividedChecksums();
    m_reset->setEnabled(true);
}

void ChecksumsDialog::resetPartialChecksums()
{
    m_firstStartRow = 0;
    m_firstRowCount = m_ranking.count();
    divideFirstPart();
    m_reset->setEnabled(false);
    m_divideFirstPart->setEnabled(true);
    m_divideSecondPart->setEnabled(true);
}

QTableWidget * ChecksumsDialog::createTableWidget()
{
    QTableWidget *tableWidget = new QTableWidget(0, 5);
    tableWidget->setEditTriggers(QAbstractItemView::NoEditTriggers);
    tableWidget->setItemDelegate(new TableDelegate(tableWidget));
    tableWidget->verticalHeader()->hide();
    tableWidget->setShowGrid(false);
    tableWidget->setWordWrap(false);
    tableWidget->setSelectionMode(QTableWidget::NoSelection);
    tableWidget->setFocusPolicy(Qt::NoFocus);
    tableWidget->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    tableWidget->setContextMenuPolicy(Qt::NoContextMenu);
    tableWidget->setHorizontalHeaderLabels(QStringList( {
        tr("Zeile"),
        db()->isPairMode() ? tr("Paar") : tr("Spieler"),
        tr("Bobbl"),
        tr("Tore"),
        tr("Gegn. Tore")
    } ));
    return tableWidget;
}

void ChecksumsDialog::updateRankingTable(QTableWidget *table, int startRow, int rowCount)
{
    table->setRowCount(rowCount);
    int color = 1;
    for (int i = startRow; i < startRow + rowCount; i++) {
        color++;
        color = color % 2;
        table->setItem(i - startRow, 0, new Cell(QString::number(i + 1), color));
        table->setItem(i - startRow, 1, new Cell(m_ranking.at(i).name, color));
        table->setItem(i - startRow, 2, new Cell(QString::number(m_ranking.at(i).boogers), color));
        table->setItem(i - startRow, 3, new Cell(QString::number(m_ranking.at(i).goals), color));
        table->setItem(i - startRow, 4, new Cell(QString::number(m_ranking.at(i).opponentGoals),
                                                 color));
    }
    table->resizeColumnsToContents();
}
