/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATABASE_H
#define DATABASE_H

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QMap>
#include <QColor>

// Local classes
class StringTools;

// Qt classes
class QSqlQuery;
class QLockFile;

class Database : public QObject
{
    Q_OBJECT

public:
    enum Error {
        NoError,
        NoErrorReadOnly,
        CouldNotOpen,
        DbVersion3,
        DbVersion4,
        DbVersion5,
        DbVersion6
    };

    enum TournamentMode {
        FixedPairs,
        SinglePlayers
    };

    enum MarkerSorting {
        NewBlock,
        Merged
    };

    struct Assignment {
        QMap<int, int> assignedTable;
        QMap<int, int> assignedPairNumber;
        QMap<int, QMap<int, int>> assignedPair;
    };

    struct Score {
        int table;
        QString pairName;
        int score;
    };

    struct SingleScore {
        int pairOrPlayer;
        int score;
    };

    struct Ranking {
        QString name;
        int boogers;
        int goals;
        int opponentGoals;
        int playedBoogers;
    };

    struct Marker {
        int id;
        QString name;
        QColor color;
        MarkerSorting sorting;
    };

    explicit Database(QObject *parent, StringTools *stringTools);
    ~Database();
    bool create(const QString &dbFile);
    Database::Error checkDatabase(const QString &dbFile);
    bool tournamentStarted() const;
    Database::Error openDb(const QString &dbFile);
    QString errorMessage() const;
    void reset();
    void saveSettings();
    void deletePlayers(const QString &name);
    void deletePlayers(const QStringList &names);
    QStringList listPlayers();
    int necessaryDivisor() const;
    void registerPlayers(const QString &name);
    void registerPlayers(const QStringList &names);
    void renamePlayers(const QString &oldName, const QString &newName);
    void renamePlayers(const QStringList &oldNames, const QStringList &newNames);
    QMap<int, bool> roundsState();
    QList<int> missingTables(int round);
    QMap<QString, int> missingPlayers(int round);
    Assignment getAssignment(int roundNumber);
    QList<Score> getScores(int round);
    QList<SingleScore> getSingleScore(int round, int table);
    QString boogify(int score);
    void saveScore(int round, int table, const QList<int> &players,
                   const QList<int> &pair1Scores, const QList<int> &pair2Scores);
    void deleteScore(int round, int table);
    QList<Ranking> getRanking(int upToRound);
    QString getMuckturnierVersion() const;
    int getDbVersion() const;
    QString getDbFile() const;
    QString getDbPath() const;
    Database::TournamentMode getTournamentMode() const;
    int getBoogerScore() const;
    int getBoogersPerRound() const;
    int playersCount();
    void setTournamentMode(Database::TournamentMode tournamentMode);
    void setBoogerScore(int boogerScore);
    void setBoogersPerRound(int boogersPerRound);
    void closeDatabaseConnection();
    void updateDatabase(const QString &dbFileName, Error dbVersion);
    void updateDatabaseFromV3(QSqlQuery &query);
    void updateDatabaseFromV4(QSqlQuery &query);
    void updateDatabaseFromV5(QSqlQuery &query);
    void updateDatabaseFromV6(QSqlQuery &query);
    bool isWritable() const;
    bool noTableNumbers() const;
    bool autoSelectPairs() const;
    bool includeOpponentGoals() const;
    void setNoTableNumbers(bool state);
    void setAutoSelectPairs(bool state);
    void setIncludeOpponentGoals(bool state);
    QList<Marker> getMarkers();
    void addMarker(const QString &name, const QColor &color, MarkerSorting sorting);
    void deleteMarker(int id, int sequence);
    void updateMarker(int id, const QString &name, const QColor &color, MarkerSorting sorting);
    void moveMarkerDown(int id, int sequence);
    void moveMarkerUp(int id, int sequence);
    void setMarker(const QString &name, int markerId);
    void removeMarker(const QString &name);
    QMap<QString, int> getPlayersMarks();
    void listSetMarker(int markerId);
    void listRemoveMarker(int markerId);
    void listRemoveAllMarkers();
    void deleteAllMarkedPlayers(int markerId);
    void deleteAllUnmarkedPlayers();
    void deleteAllPlayers();
    bool isPairMode() const;

signals:
    void sqlError(const QString &error, const QString &query = QString());

private: // Functions
    bool connectDatabase(const QString &dbFile);
    bool startTransaction();
    bool commitTransaction();
    bool queryExec(QSqlQuery &query);
    bool queryExec(QSqlQuery &query, const QString &text);
    bool queryExecBatch(QSqlQuery &query);
    void emitSqlError();
    void emitSqlError(const QSqlQuery &query);
    void updateSettings();
    void checkIfTournamentStarted();
    void insertScore(QSqlQuery &query,
                     int round, int table, int booger, int pairOrPlayer, int score);
    void updateAssignment(QSqlQuery &query,
                          int round, int pairOrPlayer, int pairNumber, int playerNumber, int table);
    bool updateValue(QSqlQuery &query, const QString &key, bool state);
    bool lockDatabase(const QString &dbFileName);
    void updateDbFileInfo();
    const QVariantList stringListToVariantList(const QStringList &stringList) const;
    void insertMarker(QSqlQuery &query, const QString &name, const QColor &color,
                      MarkerSorting sorting, int id = -1);
    void addDefaultMarkers(QSqlQuery &query);

private: // Variables
    StringTools *m_stringTools;
    QSqlDatabase m_db;
    bool m_tournamentStarted;
    QString m_errorMessage;
    QString m_dbValueMuckturnierVersion;
    TournamentMode m_settingsTournamentMode;
    int m_settingsBoogerScore;
    int m_settingsBoogersPerRound;
    bool m_isWritable;
    bool m_noTableNumbers;
    bool m_autoSelectPairs;
    bool m_includeOpponentGoals;
    QLockFile *m_lockFile;
    QString m_dbFileName;
    QString m_dbFilePath;
    bool m_dbCheckRunning;

};

#endif // DATABASE_H
