/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ExportPlayersDialog.h"
#include "SharedObjects.h"
#include "StringTools.h"
#include "Database.h"
#include "SearchEngine.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QCheckBox>
#include <QFile>
#include <QTextCodec>
#include <QApplication>
#include <QDir>
#include <QListWidget>
#include <QSplitter>
#include <QLabel>

ExportPlayersDialog::ExportPlayersDialog(QWidget *parent, SharedObjects *sharedObjects)
    : Dialog(parent, sharedObjects->database())
{
    // Dialog settings

    setTitle(db()->isPairMode() ? tr("Paarliste exportieren") : tr("Spielerliste exportieren"));
    setOkButtonText(tr("Exportieren"));

    setTargetFileBaseName(db()->isPairMode() ? tr("Paarliste") : tr("Spielerliste"));
    setTargetFileExtension(QStringLiteral("txt"));
    setTargetFileDescription(tr("Textdateien"));
    setFileDialogTitle(db()->isPairMode()
        ? tr("Bitte einen Dateinamen für die Paarliste auswählen")
        : tr("Bitte einen Dateinamen für die Spielerliste auswählen"));
    QGroupBox *selectionBox = new QGroupBox(db()->isPairMode() ? tr("Paarauswahl")
                                                               : tr("Spielerauswahl"));
    QVBoxLayout *selectionBoxLayout = new QVBoxLayout(selectionBox);
    QListWidget *markersList = new QListWidget;
    connect(markersList, &QListWidget::itemChanged, this, &ExportPlayersDialog::selectionChanged);
    selectionBoxLayout->addWidget(markersList);

    QGroupBox *previewBox = new QGroupBox(db()->isPairMode() ? tr("Exportierte Paare")
                                                             : tr("Exportierte Spieler"));
    QVBoxLayout *previewBoxLayout = new QVBoxLayout(previewBox);
    m_exportPreview = new QListWidget;
    m_exportPreview->setSelectionMode(QListWidget::NoSelection);
    m_exportPreview->setFocusPolicy(Qt::NoFocus);
    previewBoxLayout->addWidget(m_exportPreview);

    m_exportCount = new QLabel;
    previewBoxLayout->addWidget(m_exportCount);

    QSplitter *splitter = new QSplitter(this);
    splitter->setChildrenCollapsible(false);
    layout()->addWidget(splitter);
    splitter->addWidget(selectionBox);
    splitter->addWidget(previewBox);
    splitter->setStretchFactor(0, 1);
    splitter->setStretchFactor(1, 2);

    m_openAfterExport = new QCheckBox(tr("Datei nach dem Export öffnen"));
    m_openAfterExport->setChecked(true);
    layout()->addWidget(m_openAfterExport);

    // Fill the lists

    QMap<int, int> markersCount;
    const QMap<QString, int> namesAndMarks = db()->getPlayersMarks();
    QStringList names = namesAndMarks.keys();
    sharedObjects->stringTools()->sortList(names);
    for (const QString &name : names) {
        QListWidgetItem *item = new QListWidgetItem(name);
        item->setData(Qt::UserRole, namesAndMarks.value(name));
        markersCount[namesAndMarks.value(name)]++;
        m_exportPreview->addItem(item);
    }

    QMap<int, QString> markerNames;
    QMap<QString, int> markerIds;
    for (const Database::Marker &marker : db()->getMarkers()) {
        markerNames[marker.id] = marker.name;
        markerIds[marker.name] = marker.id;
    }

    for (const QString &originalName : markerIds.keys()) {
        QString name = originalName;
        if (name.isEmpty()) {
            name = tr("(unmarkiert)");
        }
        name = QStringLiteral("%1 (%2)").arg(
                              name, QString::number(markersCount.value(markerIds.value(name))));
        QListWidgetItem *item = new QListWidgetItem(name);
        item->setData(Qt::UserRole, markerIds.value(originalName));
        item->setFlags(item->flags() | Qt::ItemIsUserCheckable);
        if (markersCount.value(markerIds.value(originalName)) > 0) {
            item->setCheckState(Qt::Checked);
        } else {
            item->setCheckState(Qt::Unchecked);
            item->setFlags(item->flags() & ~Qt::ItemIsEnabled);
        }
        markersList->addItem(item);
    }

    selectionChanged(markersList->item(0));
}

void ExportPlayersDialog::selectionChanged(QListWidgetItem *changedItem)
{
    int markerId = changedItem->data(Qt::UserRole).toInt();
    bool hidden = changedItem->checkState() == Qt::Unchecked;
    int count = 0;
    for(int i = 0; i < m_exportPreview->count(); i++) {
        QListWidgetItem *nameItem = m_exportPreview->item(i);
        if (nameItem->data(Qt::UserRole).toInt() == markerId) {
            nameItem->setHidden(hidden);
            if (! hidden) {
                count++;
            }
        } else {
            if (! nameItem->isHidden()) {
                count++;
            }
        }
    }

    m_exportCount->setText(db()->isPairMode() ? tr("%n Paar(e) ausgewählt", "", count)
                                              : tr("%n Spieler ausgewählt", "", count));
    setOkButtonEnabled(count > 0);
}

void ExportPlayersDialog::accept()
{
    QString fileName = getTargetFileName(db()->getDbPath());
    if (fileName.isEmpty()) {
        return;
    }

    QFile playersListFile(fileName);
    playersListFile.open(QIODevice::WriteOnly | QIODevice::Text);

    QTextStream outStream(&playersListFile);
    outStream.setCodec(QTextCodec::codecForName("UTF-8"));

    emit updateStatus(db()->isPairMode() ? tr("Exportiere Paarliste …")
                                         : tr("Exportiere Spielerliste …")) ;
    QApplication::setOverrideCursor(Qt::WaitCursor);

    for(int i = 0; i < m_exportPreview->count(); i++) {
        QListWidgetItem *item = m_exportPreview->item(i);
        if (item->isHidden()) {
            continue;
        }
        outStream << item->text() + QStringLiteral("\n");
    }

    playersListFile.close();
    QApplication::restoreOverrideCursor();
    emit updateStatus(db()->isPairMode() ? tr("Paarliste in Datei „%2“ exportiert").arg(
                                              QDir::toNativeSeparators(fileName))
                                         : tr("Spielerliste in Datei „%2“ exportiert").arg(
                                              QDir::toNativeSeparators(fileName)));
    successMessage(db()->isPairMode() ? tr("Spielerlisten-Export") : tr("Paarlisten-Export"),
                   db()->isPairMode() ? tr("Die Paarliste wurde erfolgreich in die Datei „%2“ "
                                           "exportiert!").arg(QDir::toNativeSeparators(fileName))
                                      : tr("Die Spielerliste wurde erfolgreich in die Datei „%2“ "
                                           "exportiert!").arg(QDir::toNativeSeparators(fileName)));
    if (m_openAfterExport->isChecked()) {
        openFile(fileName);
    }

    Dialog::accept();
}
