/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ImportPlayersDialog.h"
#include "SharedObjects.h"
#include "StringTools.h"
#include "Database.h"
#include "DefaultValues.h"
#include "DefinedCharacters.h"
#include "TableDelegate.h"
#include "Cell.h"
#include "SearchEngine.h"

// Qt includes
#include <QFileDialog>
#include <QMessageBox>
#include <QTimer>
#include <QApplication>
#include <QTextStream>
#include <QVBoxLayout>
#include <QTextCodec>
#include <QGroupBox>
#include <QLabel>
#include <QTableWidget>
#include <QDebug>
#include <QHeaderView>
#include <QDebug>
#include <QScrollBar>
#include <QCheckBox>
#include <QSplitter>

ImportPlayersDialog::ImportPlayersDialog(QWidget *parent, SharedObjects *sharedObjects)
    : Dialog(parent, sharedObjects->database()), m_stringTools(sharedObjects->stringTools()),
                                                 m_searchEngine(sharedObjects->searchEngine())

{
    switch (db()->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        m_importPlayersListString = tr("Paarliste");
        break;
    case Database::TournamentMode::SinglePlayers:
        m_importPlayersListString = tr("Spielerliste");
        break;
    }
    m_title = tr("%1 importieren").arg(m_importPlayersListString),
    setTitle(m_title);

    QSplitter *splitter = new QSplitter(this);
    splitter->setOrientation(Qt::Vertical);
    splitter->setChildrenCollapsible(false);
    layout()->addWidget(splitter);

    QGroupBox *unambiguousNamesBox = new QGroupBox(db()->isPairMode() ? tr("Neue Paare")
                                                                      : tr("Neue Spieler"));
    splitter->addWidget(unambiguousNamesBox);
    QVBoxLayout *importedLayout = new QVBoxLayout(unambiguousNamesBox);

    m_unambiguousNamesLabel = new QLabel;
    m_unambiguousNamesLabel->setWordWrap(true);
    importedLayout->addWidget(m_unambiguousNamesLabel);

    m_unambiguousNames = new QTableWidget(0, 1);
    m_unambiguousNames->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_unambiguousNames->setItemDelegate(new TableDelegate(m_unambiguousNames));
    m_unambiguousNames->verticalHeader()->hide();
    m_unambiguousNames->horizontalHeader()->hide();
    m_unambiguousNames->setShowGrid(false);
    m_unambiguousNames->setWordWrap(false);
    m_unambiguousNames->setSelectionMode(QTableWidget::NoSelection);
    m_unambiguousNames->setFocusPolicy(Qt::NoFocus);
    m_unambiguousNames->setContextMenuPolicy(Qt::NoContextMenu);
    importedLayout->addWidget(m_unambiguousNames);

    m_possibleDuplicatesBox = new QGroupBox(tr("Mögliche Duplikate"));
    QVBoxLayout *duplicatesLayout = new QVBoxLayout(m_possibleDuplicatesBox);
    splitter->addWidget(m_possibleDuplicatesBox);
    m_possibleDuplicatesBox->hide();

    m_possibleDuplicatesLabel = new QLabel;
    m_possibleDuplicatesLabel->setWordWrap(true);
    duplicatesLayout->addWidget(m_possibleDuplicatesLabel);

    m_possibleDuplicates = new QTableWidget(0, 2);
    m_possibleDuplicates->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_possibleDuplicates->setItemDelegate(new TableDelegate(m_possibleDuplicates));
    m_possibleDuplicates->verticalHeader()->hide();
    m_possibleDuplicates->setShowGrid(false);
    m_possibleDuplicates->setWordWrap(false);
    m_possibleDuplicates->setSelectionMode(QTableWidget::NoSelection);
    m_possibleDuplicates->setFocusPolicy(Qt::NoFocus);
    m_possibleDuplicates->setContextMenuPolicy(Qt::NoContextMenu);
    m_possibleDuplicates->setHorizontalHeaderLabels(
        QStringList( { tr("Name aus der Import-%1").arg(m_importPlayersListString),
                       tr("Mögliches Duplikat von") } ));
    duplicatesLayout->addWidget(m_possibleDuplicates);
    m_possibleDuplicates->blockSignals(true);
    connect(m_possibleDuplicates, &QTableWidget::itemChanged,
            this, &ImportPlayersDialog::updateImportCount);

    m_duplicatesPrefix = tr("(Duplikat?)");
    m_addDuplicatesPrefix = new QCheckBox(tr("Den Präfix „%1“ hinzufügen").arg(m_duplicatesPrefix));
    m_addDuplicatesPrefix->setChecked(true);
    duplicatesLayout->addWidget(m_addDuplicatesPrefix);

    QTimer::singleShot(0, this, &ImportPlayersDialog::prepareImport);
}

void ImportPlayersDialog::prepareImport()
{
    // Parse an input file
    importFile();

    // Search for possible duplicates
    findPossibleDuplicates();

    // Display the collision-free new names
    displayUnambiguousNames();

    // Try to resize the window so that we don't have scrollbars
    resizeView();

    updateImportCount(nullptr);
    m_possibleDuplicates->blockSignals(false);
}

void ImportPlayersDialog::importFile()
{
    // Get the import file's name
    QString fileName = QFileDialog::getOpenFileName(this,
        m_title,
        db()->getDbPath(),
        tr("Textdateien") + QStringLiteral(" (*.txt);;")
        + tr("Alle Dateien") + QStringLiteral(" (*)"));

    if (fileName.isEmpty()) {
        QDialog::reject();
        return;
    }

    // Try to open it
    QFile inputFile(fileName);
    if (! inputFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QMessageBox::critical(this,
            m_title,
            tr("Es bestehen keine Leserechte für die Datei „%1“. Bitte die Zugriffsrechte "
               "überprüfen!").arg(QDir::toNativeSeparators(fileName)),
            QMessageBox::Ok);
        QDialog::reject();
        return;
    }

    // Check if it's suspiciously large (the default 9600 B are about 200 average pair names)
    QFileInfo info(fileName);
    if (info.size() > IMPORT_SIZE_WARNING) {
        if (QMessageBox::warning(this,
                m_title,
                tr("Die Datei „%1“ ist ungewöhnlich groß für eine %2. Soll sie trotzdem "
                   "importiert werden?").arg(QDir::toNativeSeparators(fileName),
                                             m_importPlayersListString),
                QMessageBox::Yes | QMessageBox::No,
                QMessageBox::No) == QMessageBox::No) {

            QDialog::reject();
            return;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit updateStatus(tr("Lese %1 …").arg(m_importPlayersListString));

    QTextStream stream(&inputFile);
    stream.setCodec(QTextCodec::codecForName("UTF-8"));

    int line = 0;
    bool askedAboutLongLines = false;
    int processed = 0;
    bool askedAboutManyLines = false;

    QStringList comparableList;
    int duplicates = 0;

    while (! stream.atEnd()) {
        line++;
        // Let's truncate lines at byte 256
        // just in case some user opens some binary file or whatnot ;-)

        // We also remove the legacy marker characters used in version 0.7.5 and 0.7.6, although
        // I don't think that anybody actually used this feature and still imports a pair/player
        // list containing those. But there we go ;-)
        const QString playersName = stream.readLine(256).simplified().remove(MARKER_CHARACTER);
        if (playersName.isEmpty()) {
            continue;
        }

        // If the file contains invalid characters, refuse to import it.
        if (playersName.contains(REPLACEMENT_CHARACTER)) {
            QString errorLine;
            if (playersName.length() > 45) {
                errorLine = playersName.left(45) + QStringLiteral("…");
            } else {
                errorLine = playersName;
            }

            QApplication::restoreOverrideCursor();
            emit updateStatus(tr("Import fehlgeschlagen!"));
            QMessageBox::critical(this,
                m_title,
                tr("<p>Die Datei „%1“ enthält ungültige Zeichen. Vermutlich ist es keine reine "
                   "Textdatei, oder sie ist nicht UTF-8-kodiert. Zeile Nr. %2 war die erste "
                   "fehlerhalt kodierte:</p><p>%3</p>"
                   "<p>Es kann eine UTF-8-kodierte Textdatei importiert werden, die pro Zeile "
                   "einen Paar- bzw. Spielernamen enthält.</p>").arg(
                   QDir::toNativeSeparators(fileName), QString::number(line), errorLine),
                   QMessageBox::Ok);
            QDialog::reject();
            return;
        }

        // Check for suspiciously long lines
        if (! askedAboutLongLines && playersName.length() > IMPORT_LINE_LENGTH_WARNING) {
            QApplication::restoreOverrideCursor();
            if (QMessageBox::warning(this,
                m_title,
                tr("Die Datei „%1“ enthält ungewöhnlich lange Zeilen. Soll sie trotzdem importiert "
                   "werden?").arg(QDir::toNativeSeparators(fileName)),
                QMessageBox::Yes | QMessageBox::No,
                QMessageBox::No) == QMessageBox::No) {

                emit updateStatus(tr("Import abgebrochen"));
                QDialog::reject();
                return;
            } else {
                askedAboutLongLines = true;
                QApplication::setOverrideCursor(Qt::WaitCursor);
            }
        }

        // Check for suspiciously many pairs/players
        if (! askedAboutManyLines && processed == IMPORT_LINES_COUNT_WARNING) {
            QApplication::restoreOverrideCursor();
            if (QMessageBox::warning(this,
                m_title,
                tr("Es wurden bereits %1 Zeilen aus der Datei „%2“ ausgelesen. Das sind "
                   "ungewöhnlich viele für eine %3. Soll trotzdem fortgefahren werden?").arg(
                   QString::number(processed),
                   QDir::toNativeSeparators(fileName),
                   m_importPlayersListString),
                QMessageBox::Yes | QMessageBox::No,
                QMessageBox::No) == QMessageBox::No) {

                inputFile.close();
                emit updateStatus(tr("Einlesen abgebrochen"));
                return;
            } else {
                QApplication::setOverrideCursor(Qt::WaitCursor);
                askedAboutManyLines = true;
            }
        }

        const QString comparableName = m_stringTools->toLower(playersName);
        if (! comparableList.contains(comparableName)) {
            comparableList.append(comparableName);
            m_unambiguousNamesList.append(playersName);
        } else {
            duplicates++;
        }

        processed++;
    }

    inputFile.close();
    m_stringTools->sortList(m_unambiguousNamesList);
    QApplication::restoreOverrideCursor();
    emit updateStatus(tr("Einlesen beendet"));

    if (duplicates == 1) {
        QMessageBox::information(this, m_title, tr("Aus dem eingelesenen Datensatz wurde ein "
                                                   "exaktes Duplikat entfernt."));
    } else if (duplicates > 1) {
        QMessageBox::information(this, m_title, tr("Aus dem eingelesenen Datensatz wurden %1 "
                                                   "exakte Duplikate entfernt.").arg(
                                                   QString::number(duplicates)));
    }
}

void ImportPlayersDialog::findPossibleDuplicates()
{
    // Search for possible duplicates inside the import list
    QMap<QString, QString> possibleDuplicates
        = m_searchEngine->checkListForPossibleDuplicates(m_unambiguousNamesList);
    const QStringList possibleImportListDuplicates = possibleDuplicates.keys();

    // Check against all registered players
    const QStringList playersList = db()->listPlayers();
    for (QString &name : m_unambiguousNamesList) {
        m_searchEngine->setSearchTerm(name);
        for (const QString &existing : playersList) {
            if (m_searchEngine->checkMatch(existing)) {
                possibleDuplicates[name] = existing;
                break;
            }
        }
    }

    const int count = possibleDuplicates.count();
    if (count > 0) {
        m_possibleDuplicatesBox->show();
        m_possibleDuplicates->setRowCount(count);
        if (count == 1) {
            m_possibleDuplicatesLabel->setText(tr("Es wurde ein Eintrag gefunden. Die markierten "
                                                  "Einträge werden (trotzdem) importiert."));
        } else {
            m_possibleDuplicatesLabel->setText(tr("Es wurden %1 Einträge gefunden. Die markierten "
                                                  "Einträge werden (trotzdem) importiert.").arg(
                                                  QString::number(count)));
        }

        int color = 1;
        int row = -1;
        QStringList duplicateKeys = possibleDuplicates.keys();
        m_stringTools->sortList(duplicateKeys);
        for (const QString &name : duplicateKeys) {
            color++;
            color = color % 2;
            row++;

            const QString lowerName = m_stringTools->toLower(name);
            const QString lowerExisting = m_stringTools->toLower(possibleDuplicates[name]);
            if (lowerName == lowerExisting) {
                Cell *cell = new Cell(name, color, Qt::red);
                cell->setCheckState(Qt::Unchecked);
                m_possibleDuplicates->setItem(row, 0, cell);
                m_possibleDuplicates->setItem(row, 1, new Cell(tr("(genau so angemeldet)"), color));
            } else {
                Cell *cell = new Cell(name, color);
                cell->setCheckState(Qt::Checked);
                m_possibleDuplicates->setItem(row, 0, cell);
                QString possibleDuplicate;
                if (possibleImportListDuplicates.contains(possibleDuplicates[name])) {
                    possibleDuplicate = tr("%1 (in Importliste)").arg(possibleDuplicates[name]);
                } else {
                    possibleDuplicate = tr("%1 (angemeldet)").arg(possibleDuplicates[name]);
                }
                m_possibleDuplicates->setItem(row, 1, new Cell(possibleDuplicate, color));
            }
            m_unambiguousNamesList.removeAt(m_unambiguousNamesList.indexOf(name));
        }
        m_possibleDuplicates->resizeColumnsToContents();
    }
}

void ImportPlayersDialog::displayUnambiguousNames()
{
    const int count = m_unambiguousNamesList.count();
    if (count == 0) {
        m_unambiguousNames->setRowCount(1);
        m_unambiguousNames->setItem(0, 0,
            new Cell(tr("(Alle Einträge sind mögliche Duplikate, s.\u2009U.)"), 0));
        m_unambiguousNamesLabel->hide();
    } else {
        m_unambiguousNames->setRowCount(count);
        int color = 1;
        int row = -1;
        for (const QString &name : m_unambiguousNamesList) {
            color++;
            color = color % 2;
            row++;
            m_unambiguousNames->setItem(row, 0, new Cell(name, color));
        }
        if (count == 1) {
            m_unambiguousNamesLabel->setText(tr("Es wurde ein Eintrag gefunden:"));
        } else {
            m_unambiguousNamesLabel->setText(tr("Es wurden %1 Einträge gefunden:").arg(
                                                QString::number(count)));
        }
    }
    m_unambiguousNames->resizeColumnsToContents();
}

void ImportPlayersDialog::resizeView()
{
    int targetWidth;
    int targetHeight;
    if (m_possibleDuplicates->isVisible()) {
        targetWidth = m_possibleDuplicates->horizontalHeader()->length()
                      + (frameGeometry().width() - m_possibleDuplicates->width());
        targetHeight = m_unambiguousNames->sizeHint().height()
                       + m_possibleDuplicates->sizeHint().height()
                       + (frameGeometry().height() - m_unambiguousNames->height()
                                                   - m_possibleDuplicates->height());
    } else {
        targetWidth = m_unambiguousNames->horizontalHeader()->length()
                      + (frameGeometry().width() - m_unambiguousNames->width());
        targetHeight = m_unambiguousNames->sizeHint().height()
                       + (frameGeometry().height() - m_unambiguousNames->height());
    }

    const QRect parentGeometry = dynamic_cast<QWidget *>(parent())->frameGeometry();

    if (targetWidth > parentGeometry.width()) {
        targetWidth = parentGeometry.width();
    }
    if (targetHeight > parentGeometry.height()) {
        targetHeight = parentGeometry.height();
    }
    if (targetWidth < 640) {
        targetWidth = 640;
    }
    if (targetHeight < 480) {
        targetHeight = 480;
    }

    if (targetWidth <= frameGeometry().width() && targetHeight <= frameGeometry().height()) {
        return;
    }

    targetWidth -= frameGeometry().width() - geometry().width();
    targetHeight -= frameGeometry().height() - geometry().height();

    setGeometry(parentGeometry.x() + (parentGeometry.width() - targetWidth) / 2,
                parentGeometry.y() + (parentGeometry.height() - targetHeight) / 2,
                targetWidth, targetHeight);
}

void ImportPlayersDialog::updateImportCount(QTableWidgetItem *)
{
    int count = 0;
    if (m_unambiguousNamesLabel->isVisible()) {
        // Otherwise, all entries are possible duplicates
        // but m_unambiguousNames contains one line with a message saying this
        count = m_unambiguousNames->rowCount();
    }

    int selectedDuplicates = 0;
    if (m_possibleDuplicates->isVisible()) {
        for (int i = 0; i < m_possibleDuplicates->rowCount(); i++) {
            if (m_possibleDuplicates->item(i, 0)->data(Qt::CheckStateRole) == Qt::Checked) {
                count++;
                selectedDuplicates++;
            }
        }
    }

    if (m_addDuplicatesPrefix->isVisible()) {
        m_addDuplicatesPrefix->setEnabled(selectedDuplicates > 0);
    }


    setButtonText(QDialogButtonBox::Ok,
                  db()->isPairMode() ? tr("%n Paar(e) importieren", "", count)
                                     : tr("%n Spieler importieren", "", count));
    setOkButtonEnabled(count > 0);
}

void ImportPlayersDialog::accept()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit updateStatus(db()->isPairMode() ? tr("Importiere Paare …") : tr("Importiere Spieler …"));

    QStringList comparableList = db()->listPlayers();
    // Add a lower-case version of all existing names with markers removed
    for (QString &name : comparableList) {
        name = m_stringTools->toLower(name);
    }
    // Add a lower-case version of the unambiguous names that will be imported with markers removed
    for (QString &name : m_unambiguousNamesList) {
        comparableList.append(name);
    }

    // Process all possible duplicates and generate non-colliding names
    if (m_possibleDuplicates->isVisible()) {
        QString nameToAdd;
        QString comparableName;
        for (int i = 0; i < m_possibleDuplicates->rowCount(); i++) {
            if (m_possibleDuplicates->item(i, 0)->data(Qt::CheckStateRole) != Qt::Checked) {
                continue;
            }

            nameToAdd = m_possibleDuplicates->item(i, 0)->text();
            if (m_addDuplicatesPrefix->isChecked()) {
                nameToAdd = m_duplicatesPrefix + QStringLiteral(" ") + nameToAdd;
            }

            // Generate a non-colliding name
            comparableName = m_stringTools->toLower(nameToAdd);
            if (comparableList.contains(comparableName)) {
                int i = 1;
                QString addition;
                do {
                    addition = QStringLiteral(" (%1)").arg(QString::number(++i));
                } while (comparableList.contains(comparableName + addition));
                comparableName += addition;
                nameToAdd += addition;
            }
            comparableList.append(comparableName);

            m_unambiguousNamesList.append(nameToAdd);
        }
    }

    db()->registerPlayers(m_unambiguousNamesList);

    QApplication::restoreOverrideCursor();
    const int imported = m_unambiguousNamesList.count();

    successMessage(m_title, db()->isPairMode() ? tr("%n Paar(e) importiert!", "", imported)
                                               : tr("%n Spieler importiert!", "", imported));
    Dialog::accept();
}
