/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "MainWindow.h"
#include "SharedObjects.h"
#include "AboutDialog.h"
#include "PlayersPage.h"
#include "ScorePage.h"
#include "RankingPage.h"
#include "ExportTournamentDialog.h"
#include "StopWatch.h"
#include "TournamentInfoDialog.h"
#include "NewTournamentDialog.h"
#include "DefaultValues.h"
#include "ExportPlayersDialog.h"
#include "DefinedCharacters.h"
#include "BackupDialog.h"
#include "ChecksumsDialog.h"
#include "ImportPlayersDialog.h"
#include "SharedStyles.h"
#include "ResourceFinder.h"

// Qt includes
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QTabWidget>
#include <QMenuBar>
#include <QAction>
#include <QApplication>
#include <QStatusBar>
#include <QMessageBox>
#include <QUrl>
#include <QDesktopServices>
#include <QFileInfo>
#include <QDebug>
#include <QFileDialog>
#include <QDir>
#include <QDockWidget>
#include <QPushButton>
#include <QTextCodec>
#include <QCloseEvent>
#include <QToolButton>
#include <QCheckBox>
#include <QRegularExpression>
#include <QTranslator>
#include <QLibraryInfo>
#include <QScreen>

MainWindow::MainWindow()
{
    setWindowFlags(windowFlags() | Qt::WindowContextHelpButtonHint);

    m_sharedObjects = new SharedObjects(this);

    m_db = m_sharedObjects->database();
    connect(m_db, &Database::sqlError, this, &MainWindow::displaySqlError);

    m_resourceFinder = m_sharedObjects->resourceFinder();

    // Install the translations

    // Nobody outside of Germany will use this, and nobody will ever translate
    // it to any other language. So let's just load the de locale stuff :-P
    QString path;

    // Add the Qt localizations
    QTranslator *qtTranslator = new QTranslator(this);
    // Look for a system-wide Qt locale file
    if (! qtTranslator->load(QStringLiteral("qt_de"),
                             QLibraryInfo::location(QLibraryInfo::TranslationsPath))) {
        // Look for a resource file
        path = m_resourceFinder->find(QStringLiteral("qt_de.qm"));
        if (! path.isEmpty()) {
            qtTranslator->load(QStringLiteral("qt_de"), QFileInfo(path).absolutePath());
        }
    }
    if (! qtTranslator->isEmpty()) {
        QCoreApplication::instance()->installTranslator(qtTranslator);
    }

    // Add the Muckturnier localizations (plural forms)
    QTranslator *mtTranslator = new QTranslator(this);
    path = m_resourceFinder->find(QStringLiteral("muckturnier_de.qm"));
    if (! path.isEmpty()) {
        mtTranslator->load(QStringLiteral("muckturnier_de"), QFileInfo(path).absolutePath());
    }
    if (! mtTranslator->isEmpty()) {
        QCoreApplication::instance()->installTranslator(mtTranslator);
    }

    // Window setup
    setWindowIcon(QIcon(m_resourceFinder->find(QStringLiteral("muckturnier.svg"))));
    setWindowTitle(tr("Muckturnier"));

    const QRect &availableGeometry = QGuiApplication::primaryScreen()->availableGeometry();
    if (availableGeometry.width() > 900 && availableGeometry.height() > 700) {
        resize(900, 700);
    } else {
        setWindowState(Qt::WindowMaximized);
    }

    m_centralWidget = new QWidget(this);
    setCentralWidget(m_centralWidget);
    QVBoxLayout *layout = new QVBoxLayout(m_centralWidget);

    // Banner

    QLabel *banner = new QLabel;
    banner->setPixmap(QIcon(m_resourceFinder->find(
                                QStringLiteral("banner.svg"))).pixmap(QSize(320, 80)));
    banner->setAlignment(Qt::AlignCenter);
    banner->setStyleSheet(QStringLiteral(STYLE_LOGO_HEADER));
    banner->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    layout->addWidget(banner);
    layout->addStretch();

    // Startup buttons

    QHBoxLayout *buttonLayout = new QHBoxLayout;
    layout->addLayout(buttonLayout);

    QToolButton *newTournamentButton = new QToolButton;
    newTournamentButton->setAutoRaise(true);
    newTournamentButton->setStyleSheet(QStringLiteral(STYLE_NO_BORDER));
    newTournamentButton->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    newTournamentButton->setIcon(QIcon(m_resourceFinder->find(QStringLiteral("new.svg"))));
    newTournamentButton->setIconSize(QSize(100, 100));
    newTournamentButton->setText(tr("Ein neues Turnier starten"));
    connect(newTournamentButton, &QToolButton::clicked, this, &MainWindow::startNewTournament);
    buttonLayout->addWidget(newTournamentButton);

    QToolButton *loadTournamentButton = new QToolButton;
    loadTournamentButton->setAutoRaise(true);
    loadTournamentButton->setStyleSheet(QStringLiteral(STYLE_NO_BORDER));
    loadTournamentButton->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    loadTournamentButton->setIcon(QIcon(m_resourceFinder->find(
                                            QStringLiteral("load.svg"))));
    loadTournamentButton->setIconSize(QSize(100, 100));
    loadTournamentButton->setText(tr("Ein Turnier öffnen"));
    connect(loadTournamentButton, &QPushButton::clicked, [this]() { loadTournament(); } );
    buttonLayout->addWidget(loadTournamentButton);

    // Footer

    QLabel *footer = new QLabel(tr("Das Open-Source-Projekt zum Auswerten von Muckturnieren"));
    footer->setAlignment(Qt::AlignCenter);
    footer->setStyleSheet(QStringLiteral(STYLE_LOGO_FOOTER));
    layout->addStretch();
    layout->addWidget(footer);

    // Dock setup

    setDockNestingEnabled(true);

    m_playersPage = new PlayersPage(this, m_sharedObjects);
    connect(m_playersPage, &PlayersPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::tournamentCanBeStarted,
            this, &MainWindow::setScoreTabEnabled);
    connect(m_playersPage, &PlayersPage::enableExport, this, &MainWindow::exportPossible);
    m_playersDock = new QDockWidget(tr("Paare"));
    m_playersDock->setWidget(m_playersPage);
    formatDockWidget(m_playersDock);
    m_playersDock->hide();

    m_scorePage = new ScorePage(this, m_sharedObjects);
    connect(m_scorePage, &ScorePage::tournamentStarted, this, &MainWindow::tournamentStarted);
    connect(m_scorePage, &ScorePage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_scorePage, &ScorePage::scoreChanged, m_playersPage, &PlayersPage::scoreChanged);
    m_scoreDock = new QDockWidget(tr("Ergebnisse"));
    m_scoreDock->setWidget(m_scorePage);
    formatDockWidget(m_scoreDock);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    m_scoreDock->hide();

    m_rankingPage = new RankingPage(this, m_sharedObjects);
    connect(m_rankingPage, &RankingPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::playersListChanged,
            m_rankingPage, &RankingPage::dataChanged);
    connect(m_scorePage, &ScorePage::scoreChanged, m_rankingPage, &RankingPage::dataChanged);
    m_rankingDock = new QDockWidget(tr("Rangliste"));
    m_rankingDock->setWidget(m_rankingPage);
    formatDockWidget(m_rankingDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
    m_rankingDock->hide();

    // Menu

    // "File"

    QMenu *fileMenu = menuBar()->addMenu(tr("Datei"));

    QAction *newTournamentAction = new QAction(tr("Neues Turnier starten"), this);
    fileMenu->addAction(newTournamentAction);
    connect(newTournamentAction, &QAction::triggered, this, &MainWindow::startNewTournament);

    QAction *loadTournamentAction = new QAction(tr("Turnier öffnen"), this);
    fileMenu->addAction(loadTournamentAction);
    connect(loadTournamentAction, &QAction::triggered, [this] { loadTournament(); });

    fileMenu->addSeparator();

    m_tournamentInfoAction = new QAction(tr("Turnierinformationen"), this);
    connect(m_tournamentInfoAction, &QAction::triggered, this, &MainWindow::showTournamentInfo);
    fileMenu->addAction(m_tournamentInfoAction);
    m_tournamentInfoAction->setEnabled(false);

    QMenu *imExportMenu = fileMenu->addMenu(tr("Import und Export"));

    m_importPlayersListAction = new QAction(tr("Paarliste importieren"), this);
    connect(m_importPlayersListAction, &QAction::triggered, this, &MainWindow::importPlayersList);
    imExportMenu->addAction(m_importPlayersListAction);
    m_importPlayersListAction->setEnabled(false);

    m_exportPlayersListAction = new QAction(tr("Paarliste exportieren"), this);
    connect(m_exportPlayersListAction, &QAction::triggered, this, &MainWindow::exportPlayersList);
    imExportMenu->addAction(m_exportPlayersListAction);
    m_exportPlayersListAction->setEnabled(false);

    imExportMenu->addSeparator();

    m_exportTournamentAction = new QAction(tr("Turnierdaten exportieren"), this);
    connect(m_rankingPage, &RankingPage::enableExport,
            m_exportTournamentAction, &QAction::setEnabled);
    connect(m_exportTournamentAction, &QAction::triggered,
            this, &MainWindow::showExportTournamentDialog);
    imExportMenu->addAction(m_exportTournamentAction);
    m_exportTournamentAction->setEnabled(false);

    fileMenu->addSeparator();

    m_closeTournamentAction = new QAction(tr("Turnier schließen"), this);
    fileMenu->addAction(m_closeTournamentAction);
    m_closeTournamentAction->setEnabled(false);
    connect(m_closeTournamentAction, &QAction::triggered, this, &MainWindow::closeTournament);

    QAction *quitAction = new QAction(tr("Beenden"), this);
    fileMenu->addAction(quitAction);
    connect(quitAction, &QAction::triggered, this, &QWidget::close);

    // "View"

    QMenu *viewMenu = menuBar()->addMenu(tr("Ansicht"));

    QAction *arrangeTabifiedAction = new QAction(tr("Seiten als Tabs"), this);
    connect(arrangeTabifiedAction, &QAction::triggered, this, &MainWindow::arrangeTabified);
    viewMenu->addAction(arrangeTabifiedAction);

    QAction *arrangeHorizontallyAction = new QAction(tr("Seiten nebeneinander"), this);
    connect(arrangeHorizontallyAction, &QAction::triggered, this, &MainWindow::arrangeHorizontally);
    viewMenu->addAction(arrangeHorizontallyAction);

    QAction *arrangeVerticallyAction = new QAction(tr("Seiten übereinander"), this);
    connect(arrangeVerticallyAction, &QAction::triggered, this, &MainWindow::arrangeVertically);
    viewMenu->addAction(arrangeVerticallyAction);

    QAction *arrangeOwnWindowAction = new QAction(tr("Seiten als einzelne Fenster"), this);
    connect(arrangeOwnWindowAction, &QAction::triggered, this, &MainWindow::arrangeOwnWindow);
    viewMenu->addAction(arrangeOwnWindowAction);

    // "Extras"

    QMenu *extrasMenu = menuBar()->addMenu(tr("Extras"));

    QAction *stopWatchAction = new QAction(tr("Stoppuhr"), this);
    extrasMenu->addAction(stopWatchAction);
    connect(stopWatchAction, &QAction::triggered, this, &MainWindow::showStopWatch);

    extrasMenu->addSeparator();

    m_showChecksumsDialogAction = new QAction(tr("Prüfsummen"), this);
    extrasMenu->addAction(m_showChecksumsDialogAction);
    connect(m_showChecksumsDialogAction, &QAction::triggered,
            this, &MainWindow::showChecksumsDialog);
    m_showChecksumsDialogAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_createDatabaseBackupAction = new QAction(tr("Backup der Datenbank anlegen"), this);
    connect(m_createDatabaseBackupAction, &QAction::triggered,
            this, &MainWindow::createDatabaseBackup);
    extrasMenu->addAction(m_createDatabaseBackupAction);
    m_createDatabaseBackupAction->setEnabled(false);

    m_restoreDatabaseBackupAction = new QAction(tr("Backup wiederherstellen"), this);
    connect(m_restoreDatabaseBackupAction, &QAction::triggered,
            this, &MainWindow::restoreDatabaseBackup);
    extrasMenu->addAction(m_restoreDatabaseBackupAction);
    m_restoreDatabaseBackupAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_resetDatabaseAction = new QAction(tr("Datenbank zurücksetzen"), this);
    extrasMenu->addAction(m_resetDatabaseAction);
    connect(m_resetDatabaseAction, &QAction::triggered, this, &MainWindow::resetDatabase);
    m_resetDatabaseAction->setEnabled(false);

    // "Help"

    QMenu *helpMenu = menuBar()->addMenu(tr("Hilfe"));

    QAction *handbookAction = new QAction(tr("Handbuch zu Muckturnier"), this);
    helpMenu->addAction(handbookAction);
    connect(handbookAction, &QAction::triggered, this, &MainWindow::showHandbook);

    QAction *aboutAction = new QAction(tr("Über Muckturnier"), this);
    helpMenu->addAction(aboutAction);
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAboutDialog);

    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::startNewTournament()
{
    NewTournamentDialog *dialog = new NewTournamentDialog(this, m_db);
    connect(dialog, &NewTournamentDialog::updateStatus, this, &MainWindow::updateStatus);
    if (! dialog->exec()) {
        return;
    }

    dbReset();
    databaseChanged();
}

void MainWindow::loadTournament(QString dbFileName)
{
    if (dbFileName.isEmpty()) {
        QString startPath = m_db->getDbPath();
        if (startPath.isEmpty()) {
            startPath = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation);
        }
        dbFileName = QFileDialog::getOpenFileName(this,
            tr("Bitte eine Turnierdatenbank auswählen"),
            startPath,
            tr("Muckturnier-Datenbanken") + QStringLiteral(" (*.mtdb);;")
            + tr("Alle Dateien") + QStringLiteral(" (*)"));
        if (dbFileName.isEmpty()) {
            return;
        }
    }

    updateStatus(tr("Öffne Turnierdatenbank …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);
    Database::Error openResult = m_db->openDb(dbFileName);
    QApplication::restoreOverrideCursor();

    switch (openResult) {
    case Database::Error::NoError:
        break;
    case Database::Error::NoErrorReadOnly:
        QMessageBox::warning(this,
            tr("Turnier öffnen"),
            tr("Die Datei „%1“ ist nicht beschreibbar. Die Turnierdatenbank kann nur angezeigt, "
               "aber nicht verändert werden.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        break;
    case Database::Error::DbVersion3:
    case Database::Error::DbVersion4:
    case Database::Error::DbVersion5:
    case Database::Error::DbVersion6:
        updateDatabase(dbFileName, openResult);
        return;
    case Database::Error::CouldNotOpen:
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p><b>Das Öffnen der Turnierdatenbank ist nicht möglich:</b></p>")
            + m_db->errorMessage(),
            QMessageBox::Ok);
        return;
    }

    databaseChanged();

    if (m_db->tournamentStarted()) {
        m_scoreDock->raise();
        m_scorePage->reload(-1);
    } else {
        // The score page could be enabled by the previously used database,
        // so disable it first
        m_scoreDock->hide();
        // Re-enable the score page if the database already contains
        // the necessary number of players/pairs
        const int playersCount = m_db->playersCount();
        if (playersCount > 0 && playersCount % m_db->necessaryDivisor() == 0) {
            setScoreTabEnabled(true);
        }
        m_playersDock->raise();
    }

    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_restoreDatabaseBackupAction->setEnabled(m_db->isWritable());

    updateStatus(tr("Turnierdatenbank %1 geöffnet").arg(QDir::toNativeSeparators(dbFileName)));
    QApplication::restoreOverrideCursor();
}

void MainWindow::updateDatabase(const QString &dbFileName, Database::Error dbVersion)
{
    QFileInfo fileInfo(dbFileName);

    if (! fileInfo.isWritable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p>Das Datenbankformat von Muckturnier hat sich geändert. Die Turnierdatenbank "
               "„%1“ hat die Revision %2, aktuell ist Revision %3.</p>"
               "<p>Um die Datei benutzen zu können, müsste ein Update der Datenbank durchgeführt "
               "werden, aber die Datei ist nicht beschreibbar.</p>"
               "<p>Bitte die Zugriffsrechte überprüfen, und die Datei erneut laden!</p>").arg(
               QDir::toNativeSeparators(dbFileName), QString::number(dbVersion),
                                                     QString::number(m_db->getDbVersion())),
            QMessageBox::Ok);
        return;
    }

    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Update des Datenbankformats"));
    messageBox.setText(tr("<p><b>Das Datenbankformat von Muckturnier hat sich geändert.</b></p>"
                          "<p>Soll die Turnierdatenbank „%1“ aktualisiert werden? Es wird dabei "
                          "ein Backup der Original-Datenbank im selben Ordner angelegt.</p>").arg(
                          QDir::toNativeSeparators(dbFileName)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
    messageBox.setDefaultButton(QMessageBox::Yes);

    if (messageBox.exec() != QMessageBox::Yes) {
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("Das Update des Datenbankformats wurde abgebrochen. Die Datenbank „%1“ kann nicht "
               "geöffnet werden.").arg(QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Datenbankupdate läuft …"));

    QString backupFileName = BackupDialog::createBackup(dbFileName, QString(),
                                                        QStringLiteral("original_dbv")
                                                        + QString::number(dbVersion));
    if (backupFileName.isEmpty()) {
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("<p><b>Die Backup-Datei konnte nicht angelegt werden.</b></p>"
               "<p>Das Update wird abgebrochen. Bitte die Zugriffsrechte (des Verzeichnisses „%2“) "
               "überprüfen!</p>").arg(
               QDir::toNativeSeparators(fileInfo.path())),
            QMessageBox::Ok);
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        return;
    }

    m_db->updateDatabase(dbFileName, dbVersion);

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Datenbankupdate beendet!"));
    QMessageBox::information(this,
        tr("Update des Datenbankformats"),
        tr("<p><b>Die Datenbank „%1“ wurde erfolgreich aktualisiert!</b></p>"
           "<p>Ein Backup der Originaldatei wurde mit dem Namen „%2“ erstellt.</p>").arg(
           QDir::toNativeSeparators(dbFileName),
           QDir::toNativeSeparators(backupFileName)),
        QMessageBox::Ok);

    loadTournament(dbFileName);
}

void MainWindow::showHandbook()
{
    // Opens the default browser to display the handbook.

    QString readmePath = m_resourceFinder->find(QStringLiteral("Readme.htm"));
    if (! readmePath.isEmpty()) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(readmePath));
        return;
    }

    QMessageBox::critical(this,
                          tr("Handbuch anzeigen"),
                          tr("<p>Das Handbuch kann nicht angezeigt werden: Die Datei „Readme.htm“ "
                             "wurde nicht gefunden.</p>"
                             "<p>Online ist unter <a href=\"http://muckturnier.org/readme/\">"
                             "http://muckturnier.org/readme/</a> das aktuelle Handbuch zu finden."
                             "</p>"),
                          QMessageBox::Ok);
}

void MainWindow::showAboutDialog()
{
    QDialog *aboutDialog = new AboutDialog(this, m_sharedObjects);
    aboutDialog->show();
}

void MainWindow::updateStatus(const QString &message)
{
    statusBar()->showMessage(message, 5000);
}

void MainWindow::databaseChanged()
{
    setWindowTitle(tr("Muckturnier: %1").arg(QFileInfo(m_db->getDbFile()).fileName()));
    m_centralWidget->hide();
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->resetMarkersVisibility();
    m_playersPage->reload();
    m_scorePage->setupGui();

    m_exportTournamentAction->setEnabled(false);
    m_closeTournamentAction->setEnabled(true);
    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_restoreDatabaseBackupAction->setEnabled(m_db->isWritable());

    checkEnableRankingPage();
    if (m_db->tournamentStarted()) {
        m_rankingPage->dataChanged();
        m_importPlayersListAction->setEnabled(false);
    } else {
        m_importPlayersListAction->setEnabled(m_db->isWritable());
    }

    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        m_exportPlayersListAction->setText(tr("Paarliste exportieren"));
        m_importPlayersListAction->setText(tr("Paarliste importieren"));
        break;
    case Database::TournamentMode::SinglePlayers:
        m_exportPlayersListAction->setText(tr("Spielerliste exportieren"));
        m_importPlayersListAction->setText(tr("Spielerliste importieren"));
        break;
    }

    m_resetDatabaseAction->setEnabled(m_db->isWritable());

    // Restore the state of the options we save in the database
    m_scorePage->setNoTableNumbers(m_db->noTableNumbers());
    m_scorePage->setAutoSelectPairs(m_db->autoSelectPairs());
    m_rankingPage->setIncludeOpponentGoals(m_db->includeOpponentGoals());

    // Be sure to have the correct pairs/players selected
    m_scorePage->checkCorrectPairSelection();

    // Clear all possibly entered searches
    m_playersPage->resetSearch();
    m_scorePage->resetSearch();
    m_rankingPage->resetSearch();
}

void MainWindow::settingsSaved()
{
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();
    m_playersDock->raise();
    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_restoreDatabaseBackupAction->setEnabled(m_db->isWritable());
}

void MainWindow::setScoreTabEnabled(bool state)
{
    m_scoreDock->setVisible(state);
    if(state == true) {
        m_scorePage->reload(-1);
    }
}

void MainWindow::tournamentStarted()
{
    setPlayersTabName();
    m_playersPage->reload();
    checkEnableRankingPage();
    m_importPlayersListAction->setEnabled(! m_db->tournamentStarted());
}

void MainWindow::setPlayersTabName()
{
    m_playersDock->setWindowTitle(m_db->isPairMode() ? tr("Paare") : tr("Spieler"));
}

void MainWindow::checkEnableRankingPage()
{
    m_rankingDock->setVisible(m_db->tournamentStarted());
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
   QMainWindow::resizeEvent(event);
   if (m_scoreDock->isVisible()) {
        m_scorePage->setScrollAreaHeight();
   }
}

void MainWindow::showExportTournamentDialog()
{
    ExportTournamentDialog *dialog= new ExportTournamentDialog(this, m_db, m_rankingPage);
    connect(dialog, &ExportTournamentDialog::updateStatus, this, &MainWindow::updateStatus);
    dialog->exec();
}

void MainWindow::showStopWatch()
{
    StopWatch *stopWatch = new StopWatch(this);
    stopWatch->show();
}

void MainWindow::showTournamentInfo()
{
    TournamentInfoDialog *dialog = new TournamentInfoDialog(this, m_db);
    dialog->exec();
}

void MainWindow::formatDockWidget(QDockWidget *dock)
{
    dock->setContextMenuPolicy(Qt::PreventContextMenu);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
    dock->setStyleSheet(QStringLiteral("QDockWidget { font-weight: bold; font-size: %1pt; }").arg(
                                       (int) double(dock->font().pointSize()) * 1.2));
    addDockWidget(Qt::TopDockWidgetArea, dock);
}

void MainWindow::closeTournament()
{
    m_db->closeDatabaseConnection();
    m_playersDock->hide();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_centralWidget->show();
    m_closeTournamentAction->setEnabled(false);
    m_tournamentInfoAction->setEnabled(false);
    m_createDatabaseBackupAction->setEnabled(false);
    m_restoreDatabaseBackupAction->setEnabled(false);
    m_resetDatabaseAction->setEnabled(false);
    m_exportPlayersListAction->setEnabled(false);
    m_exportTournamentAction->setEnabled(false);
    m_importPlayersListAction->setEnabled(false);
    setWindowTitle(tr("Muckturnier"));
    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::exportPlayersList()
{
    ExportPlayersDialog *dialog = new ExportPlayersDialog(this, m_sharedObjects);
    connect(dialog, &ExportPlayersDialog::updateStatus, this, &MainWindow::updateStatus);
    dialog->exec();
    return;
}

void MainWindow::attachDocks(Qt::DockWidgetArea area)
{
    m_playersDock->setFloating(false);
    m_scoreDock->setFloating(false);
    m_rankingDock->setFloating(false);
    addDockWidget(area, m_playersDock);
    addDockWidget(area, m_scoreDock);
    addDockWidget(area, m_rankingDock);
}

void MainWindow::arrangeTabified()
{
    attachDocks(Qt::TopDockWidgetArea);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
}

void MainWindow::arrangeHorizontally()
{
    attachDocks(Qt::TopDockWidgetArea);
    splitDockWidget(m_playersDock, m_scoreDock, Qt::Horizontal);
    splitDockWidget(m_scoreDock, m_rankingDock, Qt::Horizontal);
}

void MainWindow::arrangeVertically()
{
    attachDocks(Qt::LeftDockWidgetArea);
    splitDockWidget(m_playersDock, m_scoreDock, Qt::Vertical);
    splitDockWidget(m_scoreDock, m_rankingDock, Qt::Vertical);
}

void MainWindow::arrangeOwnWindow()
{
    m_playersDock->setFloating(true);
    m_scoreDock->setFloating(true);
    m_rankingDock->setFloating(true);
}

void MainWindow::dbReset()
{
    setPlayersTabName();
    m_playersDock->hide();
    m_playersPage->resetMarkersVisibility();
    m_playersPage->reset();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_importPlayersListAction->setEnabled(true);
    m_exportPlayersListAction->setEnabled(false);
    m_exportTournamentAction->setEnabled(false);
}

void MainWindow::resetDatabase()
{
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Warning);
    messageBox.setWindowTitle(tr("Datenbank zurücksetzen"));
    messageBox.setText(tr("Sollen wirklich alle angemeldeten Spieler und alle eingegebenen "
                          "Ergebnisse gelöscht werden?"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
    messageBox.setDefaultButton(QMessageBox::Cancel);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return;
    }

    if (createBackup->isChecked()) {
        if (! createDatabaseBackup()) {
            QMessageBox::warning(this,
                tr("Datenbank zurücksetzen"),
                tr("Das angefragte Datenbankbackup konnte nicht angelegt werden, deswegen wurde "
                   "das Zurücksetzen der Datenbank abgebrochen."),
                QMessageBox::Ok);
            return;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Setzte Turnierdatenbank zurück …"));

    m_db->reset();
    setPlayersTabName();
    dbReset();
    settingsSaved();

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Turnierdatenbank zurückgesetzt."));

    QMessageBox::information(this,
        tr("Datenbank zurücksetzen"),
        tr("Die Datenbank wurde zurückgesetzt."),
        QMessageBox::Ok
    );
}

void MainWindow::importPlayersList()
{
    ImportPlayersDialog *dialog = new ImportPlayersDialog(this, m_sharedObjects);
    connect(dialog, &ImportPlayersDialog::updateStatus, this, &MainWindow::updateStatus);

    if (! dialog->exec()) {
        updateStatus(tr("Import abgebrochen"));
    } else {
        m_playersPage->updatePlayersList();
        updateStatus(tr("Import beendet"));
    }
}

void MainWindow::displaySqlError(const QString &error, const QString &query)
{
    // This should not happen

    if (QApplication::overrideCursor() != nullptr) {
        QApplication::restoreOverrideCursor();
    }

    QString message = tr("<p>Es ist ein unerwarteter Fehler bei einer Datenbankabfrage "
                         "aufgetreten. Das sollte nicht passieren!</p>"
                         "<p>Evtl. wurde die Datenbank gelöscht oder verschoben, oder die "
                         "Zugriffsrechte wurden nach dem Öffnen verändert?</p>");
    if (! query.isEmpty()) {
        message.append(tr("<p><b>Die Anfrage war:</b></p><p>%1</p>").arg(query));
    }

    message.append(tr("<p><b>Die Fehlermeldung war:</b></p>%1").arg(error));

    QMessageBox::critical(this,
        tr("Fehler in der Datenbankkommunikation"),
        message,
        QMessageBox::Ok
    );
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    bool stopWatchRunning = false;;
    for (QWidget *stopWatch : findChildren<StopWatch *>()) {
        if (dynamic_cast<StopWatch *>(stopWatch)->isRunning()) {
            stopWatchRunning = true;
            break;
        }
    }

    if (! stopWatchRunning) {
        event->accept();
        return;
    }

    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Warning);
    messageBox.setWindowTitle(tr("Muckturnier beenden"));
    messageBox.setText(tr("<p>Es läuft mindestens eine Stoppuhr! Alle Stoppuhren werden beim "
                          "Schließen ebenfalls geschlossen, und die Zeitinformation geht "
                          "verloren!</p>"
                          "<p>Soll Muckturnier trotzdem geschlossen werden?</p>"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        event->ignore();
    } else {
        event->accept();
    }
}

bool MainWindow::createDatabaseBackup()
{
    QString dbFileName = m_db->getDbFile();
    BackupDialog::DbFileInfo info = BackupDialog::parseDbFileName(dbFileName);
    QString baseName;

    // Let's see if a backup file seems to be open
    if (! info.originalBaseName.isEmpty()) {
        QMessageBox messageBox(this);
        messageBox.setWindowTitle(tr("Backup der Datenbank anlegen"));
        messageBox.setText(tr("<p><b>Ist ein Backup offen?</b></p>"
                              "<p>Es sieht so aus, als wäre die offene Turnierdatenbank „%1“ ein "
                              "Backup von „%2.mtdb“ (erstellt %3).</p>"
                              "<p>Soll im Dateinamen des jetzt angefragten Backups das Datum und "
                              "die Zeit aktualisiert oder zusätzlich angehängt werden?</p>").arg(
                              QDir::toNativeSeparators(dbFileName),
                              QDir::toNativeSeparators(info.originalBaseName),
                              BackupDialog::getDateString(info.backupDate)));
        QAbstractButton *updateButton = messageBox.addButton(tr("Aktualisieren"),
                                                             QMessageBox::AcceptRole);
        messageBox.addButton(tr("Anhängen"), QMessageBox::RejectRole);
        QPushButton *cancelButton = messageBox.addButton(QMessageBox::Cancel);
        messageBox.exec();
        if (messageBox.clickedButton() == cancelButton) {
           return false;
        }
        if (messageBox.clickedButton() == updateButton) {
            baseName = info.path + QStringLiteral("/") + info.originalBaseName;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Backup wird angelegt …"));
    m_db->closeDatabaseConnection();
    const QString backupFileName = BackupDialog::createDatabaseBackup(dbFileName, baseName);
    bool backupFailed = backupFileName.isEmpty();
    QApplication::restoreOverrideCursor();

    if (backupFailed) {
        updateStatus(tr("Backup konnte nicht erstellt werden!"));
        QMessageBox::critical(this,
            tr("Backup der Datenbank anlegen"),
            tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
               "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ "
               "überprüfen!</p>").arg(
               QDir::toNativeSeparators(info.path)
            ),
            QMessageBox::Ok);
    } else {
        updateStatus(tr("Backup erfolgreich erstellt!"));
        QMessageBox::information(this,
            tr("Backup der Datenbank anlegen"),
            tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ "
               "erstellt.").arg(
               QDir::toNativeSeparators(backupFileName)
            ),
            QMessageBox::Ok);
    }

    m_db->openDb(dbFileName);
    return backupFailed;
}

void MainWindow::restoreDatabaseBackup()
{
    BackupDialog dialog(this, m_db->getDbFile());
    if (! dialog.exec()) {
        return;
    }

    QFileInfo backupInfo(dialog.backupToRestore());
    QFileInfo dbToOverwriteInfo(dialog.dbToOverwrite());

    if (QMessageBox::question(this,
            tr("Backup wiederherstellen"),
            tr("<p>Soll das Backup „%1“ wirklich wiederhergestellt werden?</p>"
               "<p>Die Turnierdatenbank „%2“ wird dabei überschrieben und etwaige Änderungen "
               "gehen verloren!</p>").arg(backupInfo.fileName(), dbToOverwriteInfo.fileName()),
            QMessageBox::Ok | QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Stelle Backup wieder her …"));

    if (! backupInfo.isReadable()) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Die Datei „%1“ ist nicht lesbar. Bitte die Zugriffsrechte überprüfen!</p>").arg(
               QDir::toNativeSeparators(dialog.backupToRestore())));
        return;
    }

    const QString openedDb = m_db->getDbFile();
    closeTournament();

    // Is the backup we want to restore a valid database?

    if (m_db->checkDatabase(dialog.backupToRestore()) != Database::Error::NoError) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>")
            + m_db->errorMessage()
            + tr("<p>Es wird jetzt versucht, die zuletzt geöffnete Datenbank wieder zu öffnen."
                 "</p>"));

        // Try to reload the database
        loadTournament(openedDb);
        return;
    }

    // To be sure, create a backup of the database to be overwritten

    // Get a safe destination file name
    int i = 0;
    QString tempDbBackup = dialog.dbToOverwrite() + QStringLiteral("~");
    while (QFile::exists(tempDbBackup)) {
        i++;
        tempDbBackup = dialog.dbToOverwrite() + QStringLiteral("~") + QString::number(i);
    }

    // Move the file to a backup name
    if (! QFile::rename(dialog.dbToOverwrite(), tempDbBackup)) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Die Datenbank „%1“ konnte nicht nach „%2“ umbenannt werden, daher wird der "
               "Vorgang sicherheitshalber abgebrochen.</p>"
               "<p>Es wird jetzt versucht, die zuletzt geöffnete Datenbank wieder zu öffnen."
               "</p>").arg(
               QDir::toNativeSeparators(dialog.dbToOverwrite()),
               QDir::toNativeSeparators(tempDbBackup)));

        // Try to reload the database
        loadTournament(openedDb);
        return;
    }

    // Copy the backup to restore
    if (! QFile::copy(dialog.backupToRestore(), dialog.dbToOverwrite())) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));

        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Das Backup „%1“ konnte nicht nach „%2“ kopiert werden!</p>"
               "<p>Es wird jetzt versucht, die Sicherungskopie der Datenbank wiederherzustellen "
               "und die Datenbank zu öffnen.</p>").arg(
               QDir::toNativeSeparators(dialog.backupToRestore()),
               QDir::toNativeSeparators(dialog.dbToOverwrite())));

        if (! QFile::rename(tempDbBackup, dialog.dbToOverwrite())) {
            // This one really should not happen ...
            QMessageBox::critical(this,
                tr("Backup wiederherstellen"),
                tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
                   "<p>Die Sicherungskopie „%1“ der Datenbank konnte nicht nach „%2“ umbenannt "
                   "werden.</p>"
                   "<p>Die Sicherungskopie müsste vorhanden sein und sollte manuell umbenannt "
                   "werden können, um die Datenbank wiederherzustellen. Bitte überprüfen!"
                   "</p>").arg(
                QDir::toNativeSeparators(tempDbBackup),
                QDir::toNativeSeparators(dialog.dbToOverwrite())));
            return;
        }

        loadTournament(dialog.dbToOverwrite());
        return;
    }

    // Everything is fine :-)

    loadTournament(dialog.dbToOverwrite());
    QFileInfo tempBackupInfo(tempDbBackup);
    QApplication::restoreOverrideCursor();
    updateStatus(tr("Backup erfolgreich wiederhergestellt!"));

    QMessageBox::information(this,
        tr("Backup wiederherstellen"),
        tr("Das Backup „%1“ wurde erfolgreich wiederhergestellt!").arg(backupInfo.fileName()));

    if (! QFile::remove(tempDbBackup)) {
        // This should not happen
        QMessageBox::warning(this,
            tr("Backup wiederherstellen"),
            tr("Die temporäre Sicherungskopie „%1“, die vor dem Überschreiben der Datenbank "
               "angelegt wurde, konnte nicht gelöscht werden. Bitte manuell entfernen.").arg(
               QDir::toNativeSeparators(tempDbBackup)));
    }
}

void MainWindow::exportPossible(bool state)
{
    m_exportPlayersListAction->setEnabled(state);
    m_showChecksumsDialogAction->setEnabled(state);
}

void MainWindow::showChecksumsDialog()
{
    ChecksumsDialog *dialog = new ChecksumsDialog(this, m_db);
    dialog->exec();
}
