/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "MarkersWidget.h"
#include "SharedObjects.h"
#include "ResourceFinder.h"
#include "TableDelegate.h"
#include "MarkerDialog.h"
#include "TitleMenu.h"

// Qt includes
#include <QVBoxLayout>
#include <QTableWidget>
#include <QDebug>
#include <QHeaderView>
#include <QGroupBox>
#include <QApplication>
#include <QMessageBox>
#include <QFontMetrics>
#include <QCheckBox>

MarkersWidget::MarkersWidget(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent), m_db(sharedObjects->database())
{

    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    QGroupBox *markersBox = new QGroupBox(tr("Markierungen"));
    QVBoxLayout *layout = new QVBoxLayout(markersBox);
    mainLayout->addWidget(markersBox);

    m_markers = new QTableWidget;
    m_markers->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_markers->setSelectionMode(QAbstractItemView::NoSelection);
    m_markers->setFocusPolicy(Qt::NoFocus);
    m_markers->setColumnCount(4);
    m_markers->setHorizontalHeaderLabels(QStringList({ tr("n", "Anzahl"), tr("Name"),
                                                       tr("Sortierung"), QString() }));
    m_markers->verticalHeader()->hide();
    const int iconSize = QFontMetrics(QFont(font())).boundingRect(QStringLiteral("M")).height()
#ifdef Q_OS_WIN32
                         * 1.6; // On Windows, the icon looks better if it's a bit bigger
#else
                         * 1.3; // On Linux, a bit smaller looks better
#endif
    m_markers->setIconSize(QSize(iconSize, iconSize));
    m_markers->setShowGrid(false);
    m_markers->setItemDelegate(new TableDelegate(this));
    m_markers->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_markers, &QWidget::customContextMenuRequested, this, &MarkersWidget::showContextMenu);
    connect(m_markers, &QTableWidget::cellDoubleClicked,
            this, &MarkersWidget::checkShowContextMenu);
    connect(m_markers, &QTableWidget::cellClicked, this, &MarkersWidget::changeVisibility);
    layout->addWidget(m_markers);

    m_showMarkerColumn = new QCheckBox(tr("Spalte „Markierung“ anzeigen"));
    connect(m_showMarkerColumn, &QCheckBox::stateChanged, this, &MarkersWidget::showMarkerColumn);
    layout->addWidget(m_showMarkerColumn);

    // Context menu

    m_markerMenu = new TitleMenu(this);

    m_moveUpAction = new QAction(tr("Nach oben bewegen"), this);
    connect(m_moveUpAction, &QAction::triggered, this, &MarkersWidget::moveUp);
    m_markerMenu->addAction(m_moveUpAction);

    m_moveDownAction = new QAction(tr("Nach unten bewegen"), this);
    connect(m_moveDownAction, &QAction::triggered, this, &MarkersWidget::moveDown);
    m_markerMenu->addAction(m_moveDownAction);

    m_markerMenu->addSeparator();

    m_editMarkerAction = new QAction(tr("Bearbeiten"), this);
    connect(m_editMarkerAction, &QAction::triggered, this, &MarkersWidget::editMarker);
    m_markerMenu->addAction(m_editMarkerAction);

    m_deleteMarkerAction = new QAction(tr("Löschen"), this);
    connect(m_deleteMarkerAction, &QAction::triggered, this, &MarkersWidget::deleteMarker);
    m_markerMenu->addAction(m_deleteMarkerAction);

    m_markerMenu->addSeparator();

    m_newMarkerAction = new QAction(tr("Neue Markierung"), this);
    connect(m_newMarkerAction, &QAction::triggered, this, &MarkersWidget::newMarker);
    m_markerMenu->addAction(m_newMarkerAction);

    m_addMarkerMenu = new QMenu(this);
    m_addMarkerMenu->addAction(m_newMarkerAction);

    // Hide/Show context menu

    m_hideShowMarkedMenu = new QMenu(this);

    m_showOnlyAction = new QAction(this);
    m_hideShowMarkedMenu->addAction(m_showOnlyAction);

    m_showAllAction = new QAction(tr("Alle anzeigen"), this);
    connect(m_showAllAction, &QAction::triggered, this, &MarkersWidget::showAll);
    m_hideShowMarkedMenu->addAction(m_showAllAction);

    // Icons
    m_unused = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("unused.svg")));
    m_visible = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("visible.svg")));
    m_hidden = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("hidden.svg")));
}

void MarkersWidget::reload()
{
    const QList<Database::Marker> markers = m_db->getMarkers();
    m_markers->setRowCount(markers.count());
    int row = 0;
    for (const Database::Marker &marker : markers) {
        m_markers->setItem(row, 0, new QTableWidgetItem);

        QTableWidgetItem *item = new QTableWidgetItem(marker.name);
        if (marker.name.isEmpty()) {
            item->setText(tr("(unmarkiert)"));
        }
        item->setForeground(marker.color);
        item->setData(Qt::UserRole, QVariant(marker.id));
        item->setData(Qt::UserRole + 1, QVariant(marker.sorting));
        m_markers->setItem(row, 1, item);

        switch (marker.sorting) {
        case Database::MarkerSorting::NewBlock:
            item = new QTableWidgetItem(tr("Neuer Block"));
            break;
        case Database::MarkerSorting::Merged:
            item = new QTableWidgetItem(tr("Fortlaufend"));
            break;
        }
        m_markers->setItem(row, 2, item);

        if (! m_isVisible.contains(marker.id)) {
            m_isVisible[marker.id] = true;
        }
        item = new QTableWidgetItem;
        m_markers->setItem(row, 3, item);

        row++;
    }

    m_markers->horizontalHeaderItem(0)->setWhatsThis(tr(
        "In dieser Spalte wird die Anzahl der mit der entsprechenden Markierung gekennzeichneten "
        "%1 angezeigt.").arg(m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->horizontalHeaderItem(3)->setWhatsThis(tr(
        "Durch Klicken auf das Augensymbol können alle mit der entsprechenden Markierung "
        "gekennzeichneten %1 ein- bzw. ausgeblendet werden.").arg(
        m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->resizeColumnsToContents();
    emit markersChanged();
    m_showMarkerColumn->setChecked(Qt::Unchecked);
}

void MarkersWidget::showContextMenu(const QPoint &position)
{
    int column = -1;
    QString name;
    int id;
    int row;
    int count;
    const QPoint target = m_markers->viewport()->mapToGlobal(position);

    // Collect data
    QTableWidgetItem *markerItem = m_markers->itemAt(position);
    if (markerItem != nullptr) {
        column = markerItem->column();
        row = markerItem->row();
        if (column != 1) {
            markerItem = m_markers->item(row, 1);
        }
        name = m_markers->item(row, 1)->text();
        id = m_markers->item(row, 1)->data(Qt::UserRole).toInt();
        count = m_markers->item(row, 0)->text().toInt();
    }

    // Show no context menu on the counter column
    if (column == 0) {
        return;
    }

    // Show the "Hide/Show marked entries" context menu on the "Hide/Show" column
    if (column == 3) {
        m_showAllAction->setEnabled(false);
        for (const bool &visible : m_isVisible.values()) {
            if (! visible) {
                m_showAllAction->setEnabled(true);
                break;
            }
        }

        // Set the context menu to handle the selected marker
        m_showOnlyAction->setEnabled(false);
        m_showOnlyAction->disconnect();
        m_showOnlyAction->setText(tr("Nur „%1“ anzeigen").arg(name));
        connect(m_showOnlyAction, &QAction::triggered, [this, id]() { showOnly(id); });

        // Is it meaningful to enable the action?
        if (count > 0 && count < m_allMarkersCount) {
            if (! m_isVisible[id]) {
                m_showOnlyAction->setEnabled(true);
            } else {
                for (const int &markerId : m_isVisible.keys()) {
                    if (markerId == id) {
                        continue;
                    }
                    if (m_isVisible[markerId]) {
                        m_showOnlyAction->setEnabled(true);
                        break;
                    }
                }
            }
        }

        m_hideShowMarkedMenu->exec(target);
        return;
    }

    // All other actions require write access, so return here if the database if ro
    if (! m_db->isWritable()) {
        return;
    }

    // Show the "Add marker" context menu if no item was clicked
    if (column == -1) {
        m_addMarkerMenu->exec(target);
        return;
    }

    // Otherwise, show the marker context menu

    m_markerMenu->setTitle(tr("Markierung „%1“").arg(name));
    m_selectedMarker.id = id;
    m_selectedMarker.name = name;
    m_selectedMarker.color = markerItem->foreground().color();
    m_selectedMarker.sorting = static_cast<Database::MarkerSorting>(
                                   markerItem->data(Qt::UserRole + 1).toInt());
    m_selectedMarkerSequence = markerItem->row();
    m_moveUpAction->setEnabled(m_selectedMarkerSequence > 0);
    m_moveDownAction->setEnabled(m_selectedMarkerSequence < m_markers->rowCount() - 1);
    m_editMarkerAction->setEnabled(m_selectedMarker.id != 0);
    m_deleteMarkerAction->setEnabled(m_selectedMarker.id != 0);

    m_markerMenu->exec(target);
}

void MarkersWidget::newMarker()
{
    MarkerDialog dialog(this, m_markers->palette().color(QPalette::Base), m_markers);
    QString name = tr("Neue Markierung");
    int counter = 1;
    QString newName = name;
    while (m_markers->findItems(newName, Qt::MatchFixedString).size() > 0) {
        counter++;
        newName = QStringLiteral("%1 %2").arg(name, QString::number(counter));
    }
    dialog.setName(newName);
    dialog.setTitle(tr("Neue Markierung"));
    dialog.setButtonText(QDialogButtonBox::Ok, tr("Hinzufügen"));
    if (! dialog.exec()) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lege neue Markierung an …"));
    m_db->addMarker(dialog.getName(), dialog.getColor(), dialog.getSorting());
    reload();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ hinzugefügt").arg(dialog.getName()));
}

void MarkersWidget::deleteMarker()
{
    if (QMessageBox::question(this,
            tr("Markierung löschen"),
            tr("Soll die Markierung „%1“ wirklich gelöscht werden?").arg(m_selectedMarker.name),
            QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->deleteMarker(m_selectedMarker.id, m_selectedMarkerSequence);
    m_isVisible.remove(m_selectedMarker.id);
    reload();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ gelöscht").arg(m_selectedMarker.name));
}

void MarkersWidget::editMarker()
{
    MarkerDialog dialog(this, m_markers->palette().color(QPalette::Base), m_markers);
    dialog.setTitle(tr("Markierung bearbeiten"));
    dialog.setButtonText(QDialogButtonBox::Ok, tr("Speichern"));
    dialog.setName(m_selectedMarker.name);
    dialog.setEditedName(m_selectedMarker.name);
    dialog.setColor(m_selectedMarker.color);
    dialog.setSorting(m_selectedMarker.sorting);

    if (m_selectedMarkerSequence == 0) {
        dialog.disableSortingSelection();
    }

    if (! dialog.exec()) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Speichere Änderungen an Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->updateMarker(m_selectedMarker.id, dialog.getName(), dialog.getColor(),
                       dialog.getSorting());
    reload();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Änderungen an Markierung „%1“ gespeichert").arg(m_selectedMarker.name));
}

void MarkersWidget::moveDown()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Verschiebe Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->moveMarkerDown(m_selectedMarker.id, m_selectedMarkerSequence);
    reload();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ verschoben").arg(m_selectedMarker.name));
}

void MarkersWidget::moveUp()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Verschiebe Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->moveMarkerUp(m_selectedMarker.id, m_selectedMarkerSequence);
    reload();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ verschoben").arg(m_selectedMarker.name));
}

QList<Database::Marker> MarkersWidget::getMarkers() const
{
    QList<Database::Marker> markers;
    for (int row = 0; row < m_markers->rowCount(); row++) {
        Database::Marker marker;
        QTableWidgetItem *item = m_markers->item(row, 1);
        marker.id = item->data(Qt::UserRole).toInt();
        marker.name = item->text();
        marker.color = item->foreground().color();
        marker.sorting = static_cast<Database::MarkerSorting>(item->data(Qt::UserRole + 1).toInt());
        markers.append(marker);
    }
    return markers;
}

void MarkersWidget::checkShowContextMenu(int, int column)
{
    if (column == 1 || column == 2) {
        showContextMenu(m_markers->viewport()->mapFromGlobal(QCursor::pos()));
    }
}

void MarkersWidget::changeVisibility(int row, int column)
{
    if (column != 3) {
        // Not clicked on eye symbol
        return;
    }

    if (m_markers->item(row, 0)->text().toInt() == 0) {
        // Marker not used
        return;
    }

    const int id = m_markers->item(row, 1)->data(Qt::UserRole).toInt();
    m_isVisible[id] = ! m_isVisible[id];
    m_markers->item(row, 3)->setIcon(m_isVisible[id] ? m_visible : m_hidden);
    emit visibilityChanged();
}

void MarkersWidget::resetVisibility()
{
    m_isVisible.clear();
}

bool MarkersWidget::isMarkerVisible(int id) const
{
    return m_isVisible.value(id);
}

void MarkersWidget::updateCounters(QMap<int, int> &counter)
{
    m_allMarkersCount = 0;
    int markerCount;
    for (int row = 0; row < m_markers->rowCount(); row++) {
        markerCount = counter[m_markers->item(row, 1)->data(Qt::UserRole).toInt()];
        m_markers->item(row, 0)->setText(QString::number(markerCount));
        m_allMarkersCount += markerCount;
    }
    updateVisibilityIcons();
}

void MarkersWidget::showAll()
{
    for (const int &id : m_isVisible.keys()) {
        m_isVisible[id] = true;
    }
    updateVisibilityIcons();
}

void MarkersWidget::showOnly(int idToShow)
{
    for (const int &id : m_isVisible.keys()) {
        if (id == idToShow) {
            m_isVisible[id] = true;
        } else {
            m_isVisible[id] = false;
        }
    }
    updateVisibilityIcons();
}

void MarkersWidget::updateVisibilityIcons()
{
    for (int row = 0; row < m_markers->rowCount(); row++) {
        const int id = m_markers->item(row, 1)->data(Qt::UserRole).toInt();
        if (m_markers->item(row, 0)->text().toInt() == 0) {
            m_markers->item(row, 3)->setIcon(m_unused);
        } else {
            m_markers->item(row, 3)->setIcon(m_isVisible[id] ? m_visible : m_hidden);
        }
    }
    emit visibilityChanged();
}
