/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "NewTournamentDialog.h"
#include "Database.h"
#include "MainWindow.h"
#include "DefaultValues.h"

// Qt includes
#include <QGroupBox>
#include <QLabel>
#include <QApplication>
#include <QDebug>
#include <QDir>
#include <QRadioButton>
#include <QGridLayout>
#include <QSpinBox>
#include <QMessageBox>
#include <QStandardPaths>

NewTournamentDialog::NewTournamentDialog(QWidget *parent, Database *database)
    : Dialog(parent, database)
{
    setTitle(tr("Neues Turnier starten"));
    setOkButtonText(tr("Turnier starten"));
    setTargetFileBaseName(tr("Muckturnier"));
    setTargetFileExtension(QStringLiteral("mtdb"));
    setTargetFileDescription(tr("Muckturnier-Datenbanken"));
    setFileDialogTitle(tr("Bitte einen Dateinamen für die Turnierdatenbank auswählen"));

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QGridLayout *tournamentTypeBoxLayout = new QGridLayout(tournamentTypeBox);
    layout()->addWidget(tournamentTypeBox);

    m_fixedPairs = new QRadioButton(tr("Feste Paare"));
    m_fixedPairs->setChecked(true);
    tournamentTypeBoxLayout->addWidget(m_fixedPairs, 0, 0, Qt::AlignTop);

    QLabel *fixedPairsDescription = new QLabel(tr("Es werden immer zwei Spieler als festes Paar "
                                                  "angemeldet, das das ganze Turnier zusammen "
                                                  "spielt."));
    fixedPairsDescription->setWordWrap(true);
    tournamentTypeBoxLayout->addWidget(fixedPairsDescription, 0, 1);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    tournamentTypeBoxLayout->addWidget(m_singlePlayers, 1, 0, Qt::AlignTop);

    QLabel *singlePlayersDescription = new QLabel(tr("Es werden einzelne Spieler angemeldet. Die "
                                                     "Paare werden pro Runde ausgelost."));
    singlePlayersDescription->setWordWrap(true);
    tournamentTypeBoxLayout->addWidget(singlePlayersDescription, 1, 1);

    // Tournament settings

    QGroupBox *tournamentSettingsBox = new QGroupBox(tr("Turniereinstellungen"));
    QHBoxLayout *tournamentSettingsBoxWrapperLayout = new QHBoxLayout(tournamentSettingsBox);
    QGridLayout *tournamentSettingsBoxLayout = new QGridLayout;
    tournamentSettingsBoxWrapperLayout->addLayout(tournamentSettingsBoxLayout);
    tournamentSettingsBoxWrapperLayout->addStretch();
    layout()->addWidget(tournamentSettingsBox);

    QLabel *boogerScoreLabel = new QLabel(tr("Punkte pro Bobbl:"));
    boogerScoreLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(boogerScoreLabel, 0, 0);

    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(MAXIMUM_BOOGER_SCORE);
    m_boogerScore->setValue(DEFAULT_BOOGER_SCORE);
    m_boogerScore->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(m_boogerScore, 0, 1);

    QLabel *boogersPerRoundLabel = new QLabel(tr("Bobbl pro Runde:"));
    boogersPerRoundLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(boogersPerRoundLabel, 1, 0);

    m_boogersPerRound = new QSpinBox;
    m_boogersPerRound->setMinimum(1);
    m_boogersPerRound->setMaximum(MAXIMUM_BOOGERS_PER_ROUND);
    m_boogersPerRound->setValue(DEFAULT_BOOGERS_PER_ROUND);
    m_boogersPerRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentSettingsBoxLayout->addWidget(m_boogersPerRound, 1, 1);
}

void NewTournamentDialog::accept()
{
    QString fileName = getTargetFileName(
        QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation));
    if (fileName.isEmpty()) {
        return;
    }

    emit updateStatus(tr("Lege neue Turnierdatenbank an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! db()->create(fileName)) {
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Neue Turnierdatenbank anlegen"),
            tr("<p>Das Anlegen der Turnierdatenbank „%1“ ist nicht möglich:</p><p>%2</p>").arg(
               QDir::toNativeSeparators(fileName), db()->errorMessage()),
            QMessageBox::Ok);
        return;
    }

    if (m_fixedPairs->isChecked()) {
        db()->setTournamentMode(Database::TournamentMode::FixedPairs);
    } else {
        db()->setTournamentMode(Database::TournamentMode::SinglePlayers);
        db()->setAutoSelectPairs(false);
    }
    db()->setBoogerScore(m_boogerScore->value());
    db()->setBoogersPerRound(m_boogersPerRound->value());
    db()->saveSettings();

    QApplication::restoreOverrideCursor();
    emit updateStatus(tr("Turnier „%1“ gestartet.").arg(QDir::toNativeSeparators(fileName)));

    Dialog::accept();
}
