/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "PlayersPage.h"
#include "SharedObjects.h"
#include "StringTools.h"
#include "Database.h"
#include "Cell.h"
#include "SearchWidget.h"
#include "PlayersPageDelegate.h"
#include "SearchEngine.h"
#include "TitleMenu.h"
#include "MarkersWidget.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QSplitter>

PlayersPage::PlayersPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent), m_stringTools(sharedObjects->stringTools()),
                       m_db(sharedObjects->database()),
                       m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    QWidget *registerWidget = new QWidget;
    QVBoxLayout *registerLayout = new QVBoxLayout(registerWidget);

    m_newPlayersBox = new QGroupBox;
    m_newPlayersBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    QHBoxLayout *newPlayersBoxLayout = new QHBoxLayout(m_newPlayersBox);
    registerLayout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged, [this] (const QString &text) {
            m_registerButton->setEnabled( ! text.simplified().isEmpty()); } );
    connect(m_playersName, &QLineEdit::returnPressed, this, &PlayersPage::registerPlayers);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &PlayersPage::registerPlayers);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    registerLayout->addWidget(m_playersBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    playersBoxLayout->addWidget(m_isSearchLabel);

    m_playersList = new QTableWidget(this);
    m_playersList->setColumnCount(2);
    m_playersList->setColumnHidden(1, true);
    PlayersPageDelegate *delegate = new PlayersPageDelegate(m_playersList);
    m_playersList->setItemDelegate(delegate);
    connect(delegate, &PlayersPageDelegate::editEntered, this, &PlayersPage::editEntered);
    m_playersList->setHorizontalHeaderLabels({tr("Name"), tr("Markierung")});
    m_playersList->horizontalHeader()->hide();
    m_playersList->verticalHeader()->hide();
    m_playersList->setShowGrid(false);
    m_playersList->setWordWrap(false);
    m_playersList->setSelectionMode(QTableWidget::NoSelection);
    m_playersList->setFocusPolicy(Qt::NoFocus);
    m_playersList->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    connect(m_playersList->verticalScrollBar(), &QScrollBar::valueChanged,
            m_playersList, &QTableWidget::resizeColumnsToContents);
    m_playersList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_playersList, &QTableWidget::customContextMenuRequested,
            this, &PlayersPage::showContextMenu);

    playersBoxLayout->addWidget(m_playersList);

    // Search widget
    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &PlayersPage::updateSearch);
    playersBoxLayout->addWidget(m_searchWidget);

    // Markers
    m_markersWidget = new MarkersWidget(this, sharedObjects);
    connect(m_markersWidget, &MarkersWidget::markersChanged, this, &PlayersPage::markersChanged);
    connect(m_markersWidget, &MarkersWidget::statusUpdate, this, &PlayersPage::statusUpdate);
    connect(m_markersWidget, &MarkersWidget::visibilityChanged,
            this, &PlayersPage::markersVisibilityChanged);
    connect(m_markersWidget, &MarkersWidget::showMarkerColumn,
            this, &PlayersPage::showMarkerColumn);

    // Arrange the registering part and the markers part in a splitter
    QSplitter *splitter = new QSplitter(this);
    layout->addWidget(splitter);
    splitter->addWidget(registerWidget);
    splitter->setCollapsible(0, false);
    splitter->setStretchFactor(0, 5);
    splitter->addWidget(m_markersWidget);
    splitter->setCollapsible(1, true);
    splitter->setStretchFactor(1, 3);

    // Status label
    m_stateLabel = new QLabel;
    m_stateLabel->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    layout->addWidget(m_stateLabel);

    // Context menu

    m_contextMenu = new TitleMenu(this);

    m_editAction = new QAction(this);
    connect(m_editAction, &QAction::triggered, this, &PlayersPage::editPlayers);
    m_contextMenu->addAction(m_editAction);

    m_deleteAction = new QAction(this);
    connect(m_deleteAction, &QAction::triggered, this, &PlayersPage::deletePlayers);
    m_contextMenu->addAction(m_deleteAction);

    m_contextMenu->addSeparator();

    m_setOnlyMarkerAction = new QAction(this);
    m_contextMenu->addAction(m_setOnlyMarkerAction);
    m_setOnlyMarkerAction->setVisible(false);

    m_setMarkerMenu = new QMenu(tr("Markieren als"), this);
    m_contextMenu->addMenu(m_setMarkerMenu);

    m_removeMarkerAction = new QAction(tr("Markierung entfernen"), this);
    m_contextMenu->addAction(m_removeMarkerAction);
    connect(m_removeMarkerAction, &QAction::triggered, this, &PlayersPage::removeMarker);

    // Whole list menu

    m_listMenu = new TitleMenu(tr("Ganze Liste"), this);
    m_listMenu->setTitle(tr("Ganze Liste"));
    m_contextMenu->addSeparator();
    m_contextMenu->addMenu(m_listMenu);

    m_listSetOnlyMarkerAction = new QAction(this);
    m_listMenu->addAction(m_listSetOnlyMarkerAction);

    m_listSetMarkerMenu = new QMenu(tr("Markieren als"), this);
    m_listMenu->addMenu(m_listSetMarkerMenu);

    m_listRemoveOnlyMarkerAction = new QAction(this);
    m_listMenu->addAction(m_listRemoveOnlyMarkerAction);

    m_listRemoveMarkerMenu = new QMenu(tr("Markierungen entfernen"), this);
    m_listMenu->addMenu(m_listRemoveMarkerMenu);

    m_listRemoveAllMarkersAction = new QAction(tr("Alle Markierungen entfernen"), this);
    connect(m_listRemoveAllMarkersAction, &QAction::triggered,
            this, &PlayersPage::listRemoveAllMarkers);
    m_listMenu->addAction(m_listRemoveAllMarkersAction);

    m_listMenu->addSeparator();

    m_deleteOnlyMarkedAction = new QAction(this);
    m_listMenu->addAction(m_deleteOnlyMarkedAction);

    m_deleteMarkedMenu = new QMenu(this);
    m_listMenu->addMenu(m_deleteMarkedMenu);

    m_deleteUnmarkedAction = new QAction(this);
    connect(m_deleteUnmarkedAction, &QAction::triggered, this, &PlayersPage::deleteUnmarked);
    m_listMenu->addAction(m_deleteUnmarkedAction);

    m_deleteAllAction = new QAction(this);
    connect(m_deleteAllAction, &QAction::triggered, this, &PlayersPage::deleteAll);
    m_listMenu->addAction(m_deleteAllAction);
}

void PlayersPage::reload()
{
    m_newPlayersBox->show();

    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        m_newPlayersBox->setTitle(tr("Neue Paare anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Paare"));
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Paare mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Paare löschen"));
        m_deleteAllAction->setText(tr("Alle Paare löschen"));
        break;
    case Database::TournamentMode::SinglePlayers:
        m_newPlayersBox->setTitle(tr("Neue Spieler anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Spieler"));
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Spieler mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Spieler löschen"));
        m_deleteAllAction->setText(tr("Alle Spieler löschen"));
        break;
    }

    m_newPlayersBox->setEnabled(m_db->isWritable());
    if (m_db->isWritable()) {
        m_playersList->setEditTriggers(QAbstractItemView::DoubleClicked);
    } else {
        m_playersList->setEditTriggers(QAbstractItemView::NoEditTriggers);
    }

    m_markersWidget->reload(); // This will also trigger updatePlayersList()
}

void PlayersPage::markersChanged()
{
    const QList<Database::Marker> markers = m_markersWidget->getMarkers();
    const int count = markers.count();

    // No need to process count < 2, because the actions will be hidden in this case

    if (count == 2) {
        // We have exactly one marker

        m_setOnlyMarkerAction->disconnect();
        m_listSetOnlyMarkerAction->disconnect();
        m_listRemoveOnlyMarkerAction->disconnect();
        m_deleteOnlyMarkedAction->disconnect();

        for (const Database::Marker &marker : markers) {
            if (marker.id != 0) {
                m_setOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_setOnlyMarkerAction, &QAction::triggered, [this, marker] {
                        setMarker(marker.name, marker.id); });

                m_listSetOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_listSetOnlyMarkerAction, &QAction::triggered, [this, marker] {
                        listSetMarker(marker.name, marker.id); });

                m_listRemoveOnlyMarkerAction->setText(tr("Markierung „%1“ entfernen").arg(
                                                          marker.name));
                connect(m_listRemoveOnlyMarkerAction, &QAction::triggered, [this, marker] {
                        listRemoveMarker(marker.name, marker.id); });

                m_deleteOnlyMarkedAction->setText((m_db->isPairMode()
                    ? tr("Alle als „%1“ markierten Paare löschen")
                    : tr("Alle als „%1“ markierten Spieler löschen")).arg(marker.name));
                m_deleteOnlyMarkedAction->setData(marker.id);
                connect(m_deleteOnlyMarkedAction, &QAction::triggered, [this, marker] {
                        deleteMarked(marker.name, marker.id); });

                break;
            }
        }
    }

    if (count > 2) {
        // We have multiple markers
        m_setMarkerMenu->clear();
        m_listSetMarkerMenu->clear();
        m_listRemoveMarkerMenu->clear();
        m_deleteMarkedMenu->clear();

        for (const Database::Marker &marker : markers) {
            if (marker.id == 0) {
                continue;
            }

            QAction *setMarkerAction = new QAction(marker.name, m_setMarkerMenu);
            setMarkerAction->setData(marker.id);
            connect(setMarkerAction, &QAction::triggered, [this, marker] {
                    setMarker(marker.name, marker.id); });
            m_setMarkerMenu->addAction(setMarkerAction);

            QAction *listSetMarkerAction = new QAction(marker.name, m_listSetMarkerMenu);
            listSetMarkerAction->setData(marker.id);
            connect(listSetMarkerAction, &QAction::triggered, [this, marker] {
                    listSetMarker(marker.name, marker.id); });
            m_listSetMarkerMenu->addAction(listSetMarkerAction);

            QAction *listRemoveMarkerAction = new QAction(marker.name, m_listRemoveMarkerMenu);
            listRemoveMarkerAction->setData(marker.id);
            connect(listRemoveMarkerAction, &QAction::triggered, [this, marker] {
                    listRemoveMarker(marker.name, marker.id); });
            m_listRemoveMarkerMenu->addAction(listRemoveMarkerAction);

            QAction *deleteMarkedAction = new QAction(marker.name, m_deleteMarkedMenu);
            deleteMarkedAction->setData(marker.id);
            connect(deleteMarkedAction, &QAction::triggered, [this, marker] {
                    deleteMarked(marker.name, marker.id); });
            m_deleteMarkedMenu->addAction(deleteMarkedAction);
        }
    }

    m_setOnlyMarkerAction->setVisible(count == 2);
    m_setMarkerMenu->menuAction()->setVisible(count > 2);
    m_removeMarkerAction->setVisible(count > 1);

    m_listSetOnlyMarkerAction->setVisible(count == 2);
    m_listSetMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveOnlyMarkerAction->setVisible(count == 2);
    m_listRemoveMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveAllMarkersAction->setVisible(count > 2);

    m_deleteOnlyMarkedAction->setVisible(count == 2);
    m_deleteMarkedMenu->menuAction()->setVisible(count > 2);
    m_deleteUnmarkedAction->setVisible(count > 2);

    m_searchWidget->saveSearch();
    updatePlayersList();
    m_searchWidget->restoreSearch();
}

void PlayersPage::updatePlayersList()
{
    // Create a list of all blocks we need
    QMap<int, int> targetBlock;
    QList<QStringList> block;
    QMap <int, QColor> markerColor;
    QMap <int, QString> markerName;
    for (const Database::Marker &marker : m_markersWidget->getMarkers()) {
        switch (marker.sorting) {
        case Database::MarkerSorting::NewBlock:
            block.append(QStringList());
            targetBlock[marker.id] = block.count() - 1;
            break;
        case Database::MarkerSorting::Merged:
            targetBlock[marker.id] = block.count() - 1;
            break;
        }
        if (marker.id != 0) {
            markerColor[marker.id] = marker.color;
            markerName[marker.id] = marker.name;
        }
    }

    // Fetch the players list and a list of all marked players
    QStringList allPlayers = m_db->listPlayers();
    m_stringTools->sortList(allPlayers);
    const QMap<QString, int> playersMarks = m_db->getPlayersMarks();

    // Iterate through all players, add them to their respective blocks and count the markers

    m_playersList->blockSignals(true);
    m_playersList->setRowCount(allPlayers.size());
    m_searchEngine->clearSearchCache();
    QMap<int, int> markersCount;

    for (const QString &name : allPlayers) {
        m_searchEngine->updateSearchCache(name);
        block[targetBlock[playersMarks.value(name)]].append(name);
        markersCount[playersMarks.value(name)]++;
    }

    // Display all blocks

    int playersCount = 0;
    int rowType = 1;
    for (const QStringList &list : block) {
        for (const QString &name : list) {
            rowType++;
            rowType = rowType % 2;
            Cell *nameCell = new Cell(name, rowType);
            Cell *markerCell = new Cell(QString(), rowType);
            nameCell->setData(Qt::UserRole, playersMarks.value(name));
            if (playersMarks.value(name) != 0) {
                nameCell->setForeground(markerColor.value(playersMarks.value(name)));
                markerCell->setText(markerName.value(playersMarks.value(name)));
            }
            m_playersList->setItem(playersCount, 0, nameCell);
            m_playersList->setItem(playersCount, 1, markerCell);
            playersCount++;
        }
    }
    updateVisibility();

    // Update the context menus

    if (m_listSetMarkerMenu->menuAction()->isVisible()) {
        for (QAction *action : m_listSetMarkerMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] < m_playersList->rowCount());
        }
        for (QAction *action : m_listRemoveMarkerMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] > 0);
        }
        for (QAction *action : m_deleteMarkedMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] > 0);
        }
    }

    if (markersCount.value(0) == m_playersList->rowCount()) {
        m_listRemoveOnlyMarkerAction->setEnabled(false);
        m_listRemoveMarkerMenu->setEnabled(false);
        m_listRemoveAllMarkersAction->setEnabled(false);
    } else {
        m_listRemoveOnlyMarkerAction->setEnabled(true);
        m_listRemoveMarkerMenu->setEnabled(true);
        m_listRemoveAllMarkersAction->setEnabled(true);
    }

    if (m_db->tournamentStarted() || markersCount.value(0) == m_playersList->rowCount()) {
        m_deleteOnlyMarkedAction->setEnabled(false);
        m_deleteMarkedMenu->setEnabled(false);
        m_deleteUnmarkedAction->setEnabled(false);
    } else {
        m_deleteOnlyMarkedAction->setEnabled(
            markersCount[m_deleteOnlyMarkedAction->data().toInt()] < m_playersList->rowCount());
        m_deleteMarkedMenu->setEnabled(true);
        m_deleteUnmarkedAction->setEnabled(markersCount[0] > 0);
    }

    // Update the GUI

    m_markersWidget->updateCounters(markersCount);

    m_playersList->resizeColumnsToContents();
    m_playersList->blockSignals(false);

    m_searchWidget->setVisible(playersCount > 0);
    emit enableExport(playersCount > 0);

    QString state = m_db->isPairMode() ? tr("%n Paar(e) angemeldet", "", playersCount)
                                       : tr("%n Spieler angemeldet", "", playersCount);

    bool correctNumberRegistered = playersCount % m_db->necessaryDivisor() == 0;
    if (! correctNumberRegistered) {
        state.append(QStringLiteral(" – <span style=\"color:red;\">%1</span>").arg(
                                    tr("nicht durch %1 teilbar!").arg(
                                       QString::number(m_db->necessaryDivisor()))));
    }
    m_stateLabel->setText(state);
    emit tournamentCanBeStarted(playersCount > 0 && correctNumberRegistered);

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }

    m_playersName->setFocus();
    scoreChanged();
}

void PlayersPage::scoreChanged()
{
    m_deleteAction->setEnabled(! m_db->tournamentStarted());
    m_deleteAllAction->setEnabled(! m_db->tournamentStarted());
}

void PlayersPage::reset()
{
    m_playersList->clearContents();
    m_newPlayersBox->show();
}

void PlayersPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    const QString name = m_playersName->text().simplified();

    // Check for an exact match. In this case, we can't register the pair/player
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (lowerName == m_stringTools->toLower(m_playersList->item(row, 0)->text())) {
            QMessageBox::critical(this,
                m_newPlayersBox->title(),
                QStringLiteral("<p><b>%1</b></p><p>%2</p>").arg(
                    (m_db->isPairMode()
                        ? tr("Das Paar „%1“ konnte nicht gespeichert werden")
                        : tr("Der Spieler „%1“ konnte nicht gespeichert werden")).arg(name),
                    m_db->isPairMode()
                        ? tr("Es ist bereits ein Paar mit diesem Namen angemeldet.")
                        : tr("Es ist bereits ein Spieler mit diesem Namen angemeldet.")),
                QMessageBox::Ok);
                m_playersName->selectAll();
                m_playersName->setFocus();
            return;
        }
    }

    // Search for a fuzzy match
    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_searchEngine->checkMatch(m_playersList->item(row, 0)->text())) {
            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                QStringLiteral("<p>%1</p>"
                               "<p><table>"
                               "<tr><td>%2</td><td><i>%3</i></td></tr>"
                               "<tr><td>%4</td><td><i>%5</i></td></tr>"
                               "</table></p>"
                               "<p>%6</p>").arg(
                    m_db->isPairMode()
                        ? tr("Es ist bereits ein Paar mit einem ähnlichen Namen angemeldet:")
                        : tr("Es ist bereits ein Spieler mit einem ähnlichen Namen angemeldet:"),
                    tr("Eingegeben:"), name,
                    tr("Vorhanden:"), m_playersList->item(row, 0)->text(),
                    m_db->isPairMode()
                        ? tr("Das könnte zu Verwechslungen führen. Soll das Paar trotzdem "
                             "angemeldet werden?")
                        : tr("Das könnte zu Verwechslungen führen. Soll der Spieler trotzdem "
                             "angemeldet werden?")
                ),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {

                m_playersName->selectAll();
                m_playersName->setFocus();
                return;
            }
            break;
        }
    }

    emit statusUpdate(m_db->isPairMode() ? tr("Melde Paar an …") : tr("Melde Spieler an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->registerPlayers(name);
    m_playersName->clear();
    m_playersName->setFocus();
    updatePlayersList();
    m_searchWidget->clear();

    QApplication::restoreOverrideCursor();
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ angemeldet")
                                          : tr("Spieler „%1“ angemeldet")).arg(name));

    scrollTo(name);
}

void PlayersPage::scrollTo(const QString &text)
{
    m_playersList->scrollToItem(m_playersList->findItems(text, Qt::MatchExactly).at(0),
                                QAbstractItemView::PositionAtCenter);
}

void PlayersPage::editPlayers()
{
    m_playersList->editItem(m_playersList->currentItem());
}

void PlayersPage::editEntered(const QString &originalName, const QString &name)
{
    // Check for an exact match. In this case, we can't rename the pair/player
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_playersList->item(row, 0) == m_playersList->currentItem()) {
            continue;
        }
        if (lowerName == m_stringTools->toLower(m_playersList->item(row, 0)->text())) {
            QMessageBox::critical(this,
                m_newPlayersBox->title(),
                QStringLiteral("<p><b>%1</b></p><p>%2</p>").arg(
                    (m_db->isPairMode() ? tr("Paar „%1“ konnte nicht umbenannt werden:")
                                        : tr("Spieler „%1“ konnte nicht umbenannt werden:")).arg(
                                             name),
                    tr("Der Name ist bereits vergeben.")),
                QMessageBox::Ok);
            return;
        }
    }

    // Seach for a fuzzy match
    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_playersList->item(row, 0) == m_playersList->currentItem()) {
            continue;
        }
        if (m_searchEngine->checkMatch(m_playersList->item(row, 0)->text())) {
            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                QStringLiteral("<p>%1</p>"
                               "<p><table>"
                               "<tr><td>%2</td><td><i>%3</i></td></tr>"
                               "<tr><td>%4</td><td><i>%5</i></td></tr>"
                               "</table></p>"
                               "<p>%6</p>").arg(
                    m_db->isPairMode()
                        ? tr("Es ist bereits ein Paar mit einem ähnlichen Namen angemeldet:")
                        : tr("Es ist bereits ein Spieler mit einem ähnlichen Namen angemeldet:"),
                    tr("Eingegeben:"), name,
                    tr("Vorhanden:"), m_playersList->item(row, 0)->text(),
                    m_db->isPairMode()
                        ? tr("Das könnte zu Verwechslungen führen. Soll das Paar trotzdem "
                             "umbenannt werden?")
                        : tr("Das könnte zu Verwechslungen führen. Soll der Spieler trotzdem "
                             "umbenannt werden?")),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {
                return;
            }
            break;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(m_db->isPairMode() ? tr("Benenne Paar um …") : tr("Benenne Spieler um …"));
    m_db->renamePlayers(originalName, name);
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ umbenannt zu „%2“")
                                          : tr("Spieler „%1“ umbenannt zu „%2“")).arg(
                                               originalName, name));
    QApplication::restoreOverrideCursor();
    listModified();
    scrollTo(name);
}

void PlayersPage::showContextMenu(const QPoint &point)
{
    if (! m_db->isWritable()) {
        return;
    }

    QTableWidgetItem *selectedName = m_playersList->itemAt(point);
    if (selectedName == nullptr) {
        if (m_playersList->rowCount() > 0) {
            m_listMenu->exec(m_playersList->mapToGlobal(point));
        }
        return;
    }

    m_selectedName = selectedName->text();
    const int selectedMark = selectedName->data(Qt::UserRole).toInt();
    m_contextMenu->setTitle((m_db->isPairMode() ? tr("Paar „%1“")
                                                : tr("Spieler „%1“")).arg(
                                                     m_selectedName));
    m_removeMarkerAction->setEnabled(selectedMark != 0);
    for (QAction *action : m_setMarkerMenu->actions()) {
        action->setEnabled(action->data().toInt() != selectedMark);
    }

    m_contextMenu->exec(m_playersList->mapToGlobal(point));
}

void PlayersPage::updateSearch(QString searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            m_playersList->setRowHidden(row, ! isRowVisible(row));
        }
        m_playersList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    m_searchEngine->setSearchTerm(searchText);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_searchEngine->checkMatch(m_playersList->item(row, 0)->text())) {
            if (isRowVisible(row)) {
                m_playersList->showRow(row);
            }
            matches++;
        } else {
            m_playersList->hideRow(row);
        }
    }
    m_searchWidget->updateMatches(matches);
    m_playersList->resizeColumnsToContents();

    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
    m_isSearchLabel->show();
}

void PlayersPage::listModified()
{
    emit playersListChanged();
    m_searchWidget->saveSearch();
    updatePlayersList();
    m_searchWidget->restoreSearch();
}

void PlayersPage::resetSearch()
{
    m_searchWidget->clear();
}

void PlayersPage::setMarker(const QString &markerName, int markerId)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Markiere Paar „%1“ als „%2“ …")
                                          : tr("Markiere Spieler „%1“ als „%2“ …")).arg(
                                               m_selectedName, markerName));
    m_db->setMarker(m_selectedName, markerId);
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ als „%2“ markiert")
                                          : tr("Spieler „%1“ als „%2“ markiert")).arg(
                                               m_selectedName, markerName));
    QApplication::restoreOverrideCursor();
    listModified();
    scrollTo(m_selectedName);
}

void PlayersPage::removeMarker()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Entferne Markierung von Paar „%2“ …")
                                          : tr("Entferne Markierung von Spieler „%2“ …")).arg(
                                               m_selectedName));
    m_db->removeMarker(m_selectedName);
    emit statusUpdate((m_db->isPairMode() ? tr("Markierung von Paar „%2“ entfernt")
                                          : tr("Markierung von Spieler „%2“ entfernt")).arg(
                                               m_selectedName));
    QApplication::restoreOverrideCursor();
    listModified();
    scrollTo(m_selectedName);
}

void PlayersPage::listSetMarker(const QString &markerName, int markerId)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Markiere die Paarliste als „%2“")
                                          : tr("Markiere die Spielerliste als „%2“")).arg(
                                               markerName));
    m_db->listSetMarker(markerId);
    emit statusUpdate((m_db->isPairMode() ? tr("Paarliste als „%2“ markiert")
                                          : tr("Spielerliste als „%2“ markiert")).arg(
                                               markerName));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::listRemoveMarker(const QString &markerName, int markerId)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Entferne alle „%1“-Markierungen …").arg(markerName));
    m_db->listRemoveMarker(markerId);
    emit statusUpdate(tr("Alle „%1“-Markierungen entfernt").arg(markerName));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::listRemoveAllMarkers()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Entferne alle Markierungen …"));
    m_db->listRemoveAllMarkers();
    emit statusUpdate(tr("Alle Markierungen entfernt"));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::deletePlayers()
{
    if (QMessageBox::question(this,
            m_db->isPairMode() ? tr("Paar löschen")
                               : tr("Spieler löschen"),
            (m_db->isPairMode() ? tr("Soll Paar „%1“ wirklich gelöscht werden?")
                                : tr("Soll Spieler „%1“ wirklich gelöscht werden?")).arg(
                                     m_selectedName),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Lösche Paar „%1“ …")
                                          : tr("Lösche Spieler „%1“ …")).arg(m_selectedName));
    m_db->deletePlayers(m_selectedName);
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ gelöscht")
                                          : tr("Spieler „%1“ gelöscht")).arg(m_selectedName));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::deleteMarked(const QString &markerName, int markerId)
{
    if (QMessageBox::question(this,
            m_db->isPairMode() ? tr("Markierte Paare löschen") : tr("Markierte Spieler löschen"),
            (m_db->isPairMode() ? tr("Sollen wirklich alle als „%1“ markierten Paare gelöscht "
                                     "werden?")
                                : tr("Sollen wirklich alle als „%1“ markierten Spieler gelöscht "
                                     "werden?")).arg(markerName),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Lösche alle als „%1“ markierten Paare …")
                                          : tr("Lösche alle als „%1“ markierten Spieler …")).arg(
                                               markerName));
    m_db->deleteAllMarkedPlayers(markerId);
    emit statusUpdate((m_db->isPairMode() ? tr("Alle als „%1“ markierten Paare gelöscht")
                                          : tr("Alle als „%1“ markierten Spieler gelöscht")).arg(
                                               markerName));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::deleteUnmarked()
{
    if (QMessageBox::question(this,
            m_db->isPairMode() ? tr("Unmarkierte Paare löschen")
                               : tr("Unmarkierte Spieler löschen"),
            m_db->isPairMode() ? tr("Sollen wirklich alle unmarkierten Paare gelöscht werden?")
                               : tr("Sollen wirklich alle unmarkierten Spieler gelöscht werden?"),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(m_db->isPairMode() ? tr("Lösche alle unmarkierten Paare …")
                                         : tr("Lösche alle unmarkierten Spieler …"));
    m_db->deleteAllUnmarkedPlayers();
    emit statusUpdate(m_db->isPairMode() ? tr("Alle als unmarkierten Paare gelöscht")
                                         : tr("Alle als unmarkierten Spieler gelöscht"));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::deleteAll()
{
    if (QMessageBox::question(this,
            m_db->isPairMode() ? tr("Alle Paare löschen") : tr("Alle Spieler löschen"),
            m_db->isPairMode() ? tr("Sollen wirklich alle Paare gelöscht werden?")
                               : tr("Sollen wirklich alle Spieler gelöscht werden?"),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(m_db->isPairMode() ? tr("Lösche alle Paare …") : tr("Lösche alle Spieler …"));
    m_db->deleteAllPlayers();
    emit statusUpdate(m_db->isPairMode() ? tr("Alle Paare gelöscht") : tr("Alle Spieler gelöscht"));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::resetMarkersVisibility()
{
    m_markersWidget->resetVisibility();
}

void PlayersPage::markersVisibilityChanged()
{
    m_searchWidget->saveSearch();
    updateVisibility();
    m_searchWidget->restoreSearch();
}

void PlayersPage::updateVisibility()
{
    int hidden = 0;
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        bool visible = isRowVisible(row);
        m_playersList->setRowHidden(row, ! visible);
        if (! visible) {
            hidden++;
        }
    }
    if (hidden > 0) {
        m_playersBox->setTitle(m_db->isPairMode()
                                   ? tr("Angemeldete Paare (%n ausgeblendet)", "", hidden)
                                   : tr("Angemeldete Spieler (%n ausgeblendet)", "", hidden));
    } else {
        m_playersBox->setTitle(m_db->isPairMode() ? tr("Angemeldete Paare")
                                                  : tr("Angemeldete Spieler"));
    }
    m_playersList->resizeColumnsToContents();
}

bool PlayersPage::isRowVisible(int row)
{
    return m_markersWidget->isMarkerVisible(m_playersList->item(row, 0)->data(Qt::UserRole).toInt());
}

void PlayersPage::showMarkerColumn(int state)
{
    m_playersList->horizontalHeader()->setHidden(state == Qt::Unchecked);
    m_playersList->setColumnHidden(1, state == Qt::Unchecked);
    m_playersList->resizeColumnsToContents();
}
