/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "PlayersSelect.h"
#include "PlayersSelectLineEdit.h"
#include "SharedStyles.h"
#include "SearchEngine.h"
#include "PlayersSelectPopup.h"

// Qt includes
#include <QDebug>
#include <QStringListModel>
#include <QCompleter>

PlayersSelect::PlayersSelect(QWidget *parent, SearchEngine *searchEngine, int pair)
    : QComboBox(parent), m_searchEngine(searchEngine), m_pair(pair)
{
    setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    setSizeAdjustPolicy(QComboBox::AdjustToContents);
    setEditable(true);
    setInsertPolicy(QComboBox::NoInsert);
    setCompleter(0);

    connect(this, QOverload<int>::of(&QComboBox::activated),
            [this] { finishSearch(); });
    connect(this, QOverload<int>::of(&QComboBox::currentIndexChanged),
            [this](int index) { emit selectionChanged(m_pair, index); });

    m_matches = new QStringListModel(this);
    QCompleter *completer = new QCompleter(m_matches, this);
    completer->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
    connect(completer, QOverload<const QString &>::of(&QCompleter::activated),
            this, &PlayersSelect::completerActivated);

    PlayersSelectPopup *popup = new PlayersSelectPopup;
    completer->setPopup(popup);
    connect(popup, &PlayersSelectPopup::popupHidden, this, &PlayersSelect::popupHidden);

    PlayersSelectLineEdit *lineEdit = new PlayersSelectLineEdit;
    lineEdit->setCompleter(completer);
    setLineEdit(lineEdit);

    disconnect(lineEdit, &QLineEdit::editingFinished, 0, 0);
    connect(lineEdit, &QLineEdit::textEdited, this, &PlayersSelect::lineEditTextChanged);
    connect(lineEdit, &PlayersSelectLineEdit::escapePressed,
            this, &PlayersSelect::resetLineEditText);
    connect(lineEdit, &QLineEdit::editingFinished, this, &PlayersSelect::processGivenText);

    m_isSearching = false;
}

void PlayersSelect::lineEditTextChanged(const QString &text)
{
    if (! m_isSearching) {
        m_isSearching = true;
        m_completerActivated = false;
        lineEdit()->setStyleSheet(QStringLiteral(STYLE_RED_TEXT));
        emit disableSaveScore();
    }

    QStringList matches;
    m_searchEngine->setSearchTerm(text);
    for(int i = 0; i < model()->rowCount(); i++) {
        const QString name = model()->index(i, 0).data().toString();
        if (m_searchEngine->checkMatch(name)) {
            matches.append(name);
        }
    }
    m_matches->setStringList(matches);
}

void PlayersSelect::resetLineEditText()
{
    lineEdit()->setText(itemText(currentIndex()));
    finishSearch();
    lineEdit()->setFocus();
}

void PlayersSelect::processGivenText()
{
    if (! m_isSearching || m_completerActivated) {
        return;
    }

    if (m_matches->rowCount() == 1) {
        completerActivated(m_matches->index(0, 0).data().toString());
        return;
    }

    resetLineEditText();
}

void PlayersSelect::completerActivated(const QString &text)
{
    m_completerActivated = true;
    setCurrentIndex(findText(text));
    finishSearch();
}

void PlayersSelect::finishSearch()
{
    if (m_isSearching) {
        m_isSearching = false;
        lineEdit()->setStyleSheet(QString());
    }
    emit checkSavePossible();
}

void PlayersSelect::popupHidden()
{
    if (m_completerActivated || m_matches->rowCount() != 1) {
        return;
    }

    const QString &onlyMatch = m_matches->index(0, 0).data().toString();
    if (lineEdit()->text() != onlyMatch) {
        completerActivated(onlyMatch);
    }
}
