/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ScoreDialog.h"

// Qt includes
#include <QDebug>
#include <QGridLayout>
#include <QPushButton>
#include <QSpinBox>
#include <QCursor>
#include <QScrollArea>
#include <QScrollBar>
#include <QApplication>
#include <QTimer>
#include <QScreen>

ScoreDialog::ScoreDialog(QWidget *parent) : QDialog(parent)
{
    setWindowFlags(Qt::FramelessWindowHint | Qt::Popup);
    setStyleSheet(QStringLiteral(
        "QDialog { border: 1px solid gray; } "
        "QScrollArea { border: none; } "
        "QPushButton { border: none; border-radius: 4px; padding:0.5em; } "
        "QPushButton:hover { border: 1px solid palette(highlight); } "
    ));

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->setSpacing(0);
    layout->setMargin(1); // We need this for the QDialog's gray border

    m_scrollArea = new QScrollArea(this);
    m_scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    layout->addWidget(m_scrollArea);

    m_boogerScore = -1;
}

void ScoreDialog::setBoogerScore(int score)
{
    if (m_boogerScore == score) {
        return;
    }

    QWidget *scoreWidget = new QWidget;
    scoreWidget->setStyleSheet(QStringLiteral("background-color: white;"));
    QGridLayout *layout = new QGridLayout(scoreWidget);

    int maximumColumns;
    if (score <= 31) {
        maximumColumns = 5;
    } else {
        maximumColumns = 10;
    }

    int row = -1;
    int currentScore = -1;
    bool finished = false;
    while (! finished) {
        row++;
        for (int column = 0; column < maximumColumns; column++) {
            currentScore++;
            if (currentScore == score) {
                finished = true;
                break;
            }

            QPushButton *scoreButton = new QPushButton(QString::number(currentScore));
            scoreButton->setProperty("score", QVariant(currentScore));
            layout->addWidget(scoreButton, row, column);
            connect(scoreButton, &QPushButton::clicked,
                    [this, scoreButton] { scoreClicked(scoreButton); });
        }
    }

    m_scrollArea->setWidget(scoreWidget);
    m_boogerScore = score;
}

void ScoreDialog::show()
{
    const int availableHeight = QGuiApplication::primaryScreen()->availableGeometry().height()
                                - QCursor::pos().y();
    const int widgetHeight    = m_scrollArea->widget()->size().height()
                                + 2; // Two pixels for the border

    if (widgetHeight <= availableHeight) {
        resize(m_scrollArea->widget()->width(), widgetHeight);
    } else {
        resize(m_scrollArea->widget()->width(), availableHeight);
    }

    move(QCursor::pos());
    QDialog::show();
    QTimer::singleShot(0, this, &ScoreDialog::checkForScrollBar);
}

void ScoreDialog::checkForScrollBar()
{
    if (m_scrollArea->verticalScrollBar()->isVisible()) {
        resize(m_scrollArea->widget()->width() + m_scrollArea->verticalScrollBar()->width(),
               height());
    } else {
        resize(m_scrollArea->widget()->width(), height());
    }
}

void ScoreDialog::scoreClicked(QPushButton *button)
{
    m_score = button->property("score").toInt();
    // This is needed to prevent the button to be highlighted afterwards on Windows
    button->clearFocus();
    accept();
}

int ScoreDialog::score() const
{
    return m_score;
}
