/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ScorePage.h"
#include "SharedObjects.h"
#include "StringTools.h"
#include "BoogerWidget.h"
#include "Cell.h"
#include "SearchWidget.h"
#include "TableDelegate.h"
#include "SharedStyles.h"
#include "SearchEngine.h"
#include "TitleMenu.h"
#include "PlayersSelect.h"
#include "BoogerSpinBox.h"
#include "ScoreDialog.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QComboBox>
#include <QGridLayout>
#include <QCheckBox>
#include <QScrollArea>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QAction>
#include <QHBoxLayout>
#include <QMessageBox>
#include <QApplication>
#include <QScrollBar>
#include <QLabel>

ScorePage::ScorePage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent), m_stringTools(sharedObjects->stringTools()),
                       m_db(sharedObjects->database()),
                       m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Round

    QGroupBox *roundBox = new QGroupBox;
    QHBoxLayout *roundBoxLayout = new QHBoxLayout(roundBox);
    layout->addWidget(roundBox);

    QLabel *roundLabel = new QLabel(tr("Runde:"));
    roundLabel->setStyleSheet(QStringLiteral(STYLE_BOLD_TEXT));
    roundLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    roundBoxLayout->addWidget(roundLabel);

    m_roundSelect = new QComboBox;
    m_roundSelect->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_roundSelect->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    roundBoxLayout->addWidget(m_roundSelect);
    connect(m_roundSelect, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &ScorePage::openRound);

    roundBoxLayout->addStretch();

    m_noTableNumbers = new QCheckBox(tr("Eingabe ohne Tischnummern"));
    m_noTableNumbers->setWhatsThis(tr("Wenn ohne Tischnummern gespielt wird, dann wird die "
                                      "Tischauswahl und -anzeige ausgeblendet, wenn diese Option "
                                      "gesetzt ist."));
    roundBoxLayout->addWidget(m_noTableNumbers);
    connect(m_noTableNumbers, &QCheckBox::stateChanged, this, &ScorePage::noTableNumbersChanged);

    m_resultBox = new QGroupBox(tr("Ergebnisse eintragen"));
    QVBoxLayout *resultBoxLayout = new QVBoxLayout(m_resultBox);
    layout->addWidget(m_resultBox);

    // Table and pairs/players

    QGroupBox *playersBox = new QGroupBox;
    QGridLayout *playersBoxLayout = new QGridLayout(playersBox);
    resultBoxLayout->addWidget(playersBox);

    m_tableLabel = new QLabel(tr("Tisch:"));
    m_tableLabel->setStyleSheet(QStringLiteral(STYLE_BOLD_TEXT));
    m_tableLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    playersBoxLayout->addWidget(m_tableLabel, 0, 0, 2, 1);

    m_tableSelect = new QComboBox;
    m_tableSelect->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_tableSelect->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    playersBoxLayout->addWidget(m_tableSelect, 0, 1, 2, 1);
    connect(m_tableSelect, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &ScorePage::tableChanged);

    for (int pair = 1; pair <= 2; pair++) {
        QLabel *playersLabel = new QLabel(tr("Paar %1:").arg(QString::number(pair)));
        playersLabel->setStyleSheet(QStringLiteral(STYLE_BOLD_TEXT));
        playersLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
        playersBoxLayout->addWidget(playersLabel, pair - 1, 2);

        QMap<int, PlayersSelect *> entry;
        for (int pairPart = 1; pairPart <= 2; pairPart++) {
            PlayersSelect *pairPartComboBox = new PlayersSelect(this, m_searchEngine, pair);
            connect(pairPartComboBox, &PlayersSelect::selectionChanged,
                    this, &ScorePage::playersChanged);
            connect(pairPartComboBox, &PlayersSelect::disableSaveScore,
                    this, &ScorePage::disableSaveScore);
            connect(pairPartComboBox, &PlayersSelect::checkSavePossible,
                    this, &ScorePage::checkSavePossible);
            playersBoxLayout->addWidget(pairPartComboBox, pair - 1, 2 + pairPart);
            entry[pairPart] = pairPartComboBox;
        }
        m_playersSelect[pair] = entry;
    }

    m_autoSelectPairs = new QCheckBox(tr("Tischnummer bzw. Paare\nautomatisch auswählen"));
    m_autoSelectPairs->setWhatsThis(tr(
        "<p>Wird nach dem System „Paar 1 bleibt sitzen, Paar 2 rückt jede Runde einen Tisch "
        "weiter“ gespielt, dann werden mittels dieser Option die Paare für einen Tisch oder der "
        "Tisch und das gegnerische Paar für ein anderes Paar automatisch ausgewählt.</p>"
        "<p>Hierzu muss natürlich die Auslosung bekannt sein, so dass die automatische Auswahl "
        "erst ab Runde 2 funktionieren kann. Am einfachsten ist die Auswahl über die Tischnummer!"
        "</p>"
    ));
    playersBoxLayout->addWidget(m_autoSelectPairs, 0, 5, 2, 1);
    connect(m_autoSelectPairs, &QCheckBox::stateChanged, this, &ScorePage::autoSelectPairsChanged);

    // This works as a spacer so that the place is filled if m_autoSelectPairs is not shown
    playersBoxLayout->addWidget(new QLabel, 0, 6);

    // Enter Score
    QGroupBox *scoreBox = new QGroupBox;
    QVBoxLayout *scoreBoxLayout = new QVBoxLayout(scoreBox);
    scoreBox->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    resultBoxLayout->addWidget(scoreBox);

    m_scoreBoxScroll = new QScrollArea(scoreBox);
    m_scoreBoxScroll->setFrameShape(QFrame::NoFrame);
    scoreBoxLayout->addWidget(m_scoreBoxScroll);

    m_saveScoreButton = new QPushButton(tr("Ergebnis\neintragen"));
    connect(m_saveScoreButton, &QPushButton::clicked, this, &ScorePage::saveScore);

    // The round's results

    m_roundSummaryBox = new QGroupBox;
    QVBoxLayout *roundSummaryBoxLayout = new QVBoxLayout(m_roundSummaryBox);
    layout->addWidget(m_roundSummaryBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    roundSummaryBoxLayout->addWidget(m_isSearchLabel);

    m_scoreList = new QTableWidget;
    m_scoreList->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_scoreList->setItemDelegate(new TableDelegate(this, TableDelegate::HighlightMode::EvenRows));
    m_scoreList->verticalHeader()->hide();
    m_scoreList->setShowGrid(false);
    m_scoreList->setWordWrap(false);
    m_scoreList->setSelectionMode(QTableWidget::NoSelection);
    m_scoreList->setFocusPolicy(Qt::NoFocus);
    m_scoreList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_scoreList, &QTableWidget::cellDoubleClicked,
            [this] { showContextMenu(m_scoreList->viewport()->mapFromGlobal(QCursor::pos())); } );
    connect(m_scoreList, &QTableWidget::customContextMenuRequested,
            this, &ScorePage::showContextMenu);
    roundSummaryBoxLayout->addWidget(m_scoreList);

    // Search widget

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &ScorePage::updateSearch);
    roundSummaryBoxLayout->addWidget(m_searchWidget);

    // Context menu

    m_contextMenu = new TitleMenu(this);

    m_editAction = new QAction(tr("Ändern (Löschen und neu eingeben)"), this);
    m_contextMenu->addAction(m_editAction);
    connect(m_editAction, &QAction::triggered, this, &ScorePage::editScore);

    m_swapPairsAction = new QAction(tr("Paar 1 und Paar 2 vertauschen"), this);
    m_contextMenu->addAction(m_swapPairsAction);
    connect(m_swapPairsAction, &QAction::triggered, this, &ScorePage::swapPairs);

    m_contextMenu->addSeparator();

    m_deleteAction = new QAction(tr("Löschen"), this);
    m_contextMenu->addAction(m_deleteAction);
    connect(m_deleteAction, &QAction::triggered, this, &ScorePage::deleteScore);

    // Score dialog
    m_scoreDialog = new ScoreDialog(this);
    connect(m_scoreDialog, &QDialog::accepted, this, &ScorePage::setSelectedScore);
}

void ScorePage::setupGui()
{
    // Show the needed input fields for pairs or players
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        m_playersSelect[1][2]->hide();
        m_playersSelect[2][2]->hide();
        m_autoSelectPairs->show();
        break;
    case Database::TournamentMode::SinglePlayers:
        m_playersSelect[1][2]->show();
        m_playersSelect[2][2]->show();
        m_autoSelectPairs->hide();
        break;
    }

    // Display the correct number of booger widgets

    m_boogerWidgets.clear();
    QWidget *scoreBoxWidget = new QWidget;
    QHBoxLayout *scoreBoxWidgetLayout = new QHBoxLayout(scoreBoxWidget);
    scoreBoxWidgetLayout->setSizeConstraint(QLayout::SetFixedSize);

    for (int i = 1; i <= m_db->getBoogersPerRound(); i++) {
        BoogerWidget *boogerWidget = new BoogerWidget(scoreBoxWidget, i, m_db->getBoogersPerRound(),
                                                      m_db->getBoogerScore());
        connect(boogerWidget, &BoogerWidget::abortedStateChanged,
                this, &ScorePage::boogerAbortedStateChanged);
        connect(boogerWidget, &BoogerWidget::scoreDialogRequested,
                this, &ScorePage::showScoreDialog);

        m_boogerWidgets.append(boogerWidget);
        scoreBoxWidgetLayout->addWidget(boogerWidget);
    }
    connect(m_boogerWidgets.last(), &BoogerWidget::returnPressed,
            this, &ScorePage::checkEnterScore);

    // Add the "Save score" button
    scoreBoxWidgetLayout->addWidget(m_saveScoreButton);
    m_saveScoreButton->setEnabled(false);
    m_scoreBoxScroll->setWidget(scoreBoxWidget);
    scoreBoxWidget->setAutoFillBackground(false);
    setScrollAreaHeight();

    // Disable data input if we have a read-only database
    m_resultBox->setEnabled(m_db->isWritable());

    // Set the correct score widget's column count
    m_scoreList->setColumnCount(m_db->getBoogersPerRound() + 2);
    QStringList headerLabels({ tr("Tisch"), tr("Paare") });
    for (int i = 1; i <= m_db->getBoogersPerRound(); i++) {
        headerLabels.append(tr("%1. Bobbl").arg(QString::number(i)));
    }
    m_scoreList->setHorizontalHeaderLabels(headerLabels);

    // Update the score dialog
    m_scoreDialog->setBoogerScore(m_db->getBoogerScore());
}

void ScorePage::reload(int lastRound)
{
    // Reset all booger widgets if the GUI has been already setup
    if (lastRound != -1) {
        for (BoogerWidget *boogerWidget : m_boogerWidgets) {
            boogerWidget->reset();
        }
        m_boogerWidgets.at(0)->setFocus();
    }

    // Fill the round select

    blockSelectSignals(true);

    int allRounds = 0;
    int firstOpenRound = -1;

    m_roundSelect->clear();
    m_roundsState = m_db->roundsState();
    for (const int &roundNumber : m_roundsState.keys()) {
        if (m_roundsState.value(roundNumber)) {
            m_roundSelect->addItem(tr("%1 (abgeschlossen)").arg(QString::number(roundNumber)),
                                   roundNumber);
        } else {
            m_roundSelect->addItem(tr("%1 (offen)").arg(QString::number(roundNumber)),
                                   roundNumber);
            if (firstOpenRound == -1) {
                firstOpenRound = m_roundSelect->count() - 1;
            }
        }
        allRounds++;
    }

    if (firstOpenRound == -1) {
        firstOpenRound = 0;
    }
    if (lastRound != -1) {
        firstOpenRound = lastRound - 1;
    }

    // Add one for a new round
    if (m_db->isWritable()) {
        allRounds++;
        m_roundSelect->addItem(tr("%1 (neue Runde)").arg(QString::number(allRounds)), allRounds);
        m_roundsState[allRounds] = false;
    }

    // Display the data
    m_roundSelect->setCurrentIndex(firstOpenRound);
    blockSelectSignals(false);
    openRound(firstOpenRound);
}

void ScorePage::setScrollAreaHeight()
{
    if (m_scoreBoxScroll->widget() == nullptr) {
        // This happens when no database has been created/loaded yet
        return;
    }

    if (m_scoreBoxScroll->width() < m_scoreBoxScroll->widget()->width()) {
        m_scoreBoxScroll->setMinimumHeight(m_scoreBoxScroll->widget()->height()
                                           + m_scoreBoxScroll->horizontalScrollBar()->height());
    } else {
        m_scoreBoxScroll->setMinimumHeight(m_scoreBoxScroll->widget()->height());
    }
}

void ScorePage::openRound(int roundNumber)
{
    if (roundNumber == -1) {
        return;
    }

    blockSelectSignals(true);

    m_searchWidget->saveSearch();
    int round = m_roundSelect->itemData(roundNumber).toInt();
    m_resultBox->setVisible(! m_roundsState.value(round));

    // Insert all missing tables
    m_tableSelect->clear();
    for (const int &table : m_db->missingTables(round)) {
        m_tableSelect->addItem(QString::number(table), table);
    }

    // Insert all missing players/pairs

    for (int i = 1; i <= 2; i++) {
        m_playersSelect[i][1]->clear();
        if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
            m_playersSelect[i][2]->clear();
        }
    }

    QMap<QString, int> missingPlayers = m_db->missingPlayers(round);
    m_availableNames = missingPlayers.keys();
    m_stringTools->sortList(m_availableNames);
    for (const QString &missingPlayersName : m_availableNames) {
        for (int i = 1; i <= 2; i++) {
            m_playersSelect[i][1]->addItem(missingPlayersName,
                                           missingPlayers.value(missingPlayersName));
            if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
                m_playersSelect[i][2]->addItem(missingPlayersName,
                                               missingPlayers.value(missingPlayersName));
            }
        }
    }

    updateScore();

    m_saveScoreButton->setEnabled(false);

    m_assignment = m_db->getAssignment(round);
    blockSelectSignals(false);
    setScrollAreaHeight();
    tableChanged(0);
    checkSavePossible();
    m_searchWidget->restoreSearch();
}

void ScorePage::updateScore()
{
    const QList<Database::Score> score = m_db->getScores(
        m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt());
    m_scoreList->setRowCount(score.size() / m_db->getBoogersPerRound());

    if (score.size() == 0) {
        m_searchWidget->clear();
        m_isSearchLabel->hide();
        m_searchWidget->hide();
    } else {
        m_searchWidget->show();
    }

    if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
        m_searchEngine->clearSearchCache(SearchEngine::SearchType::DrawnPairsScore);
    }

    int color = 1;
    int row = -1;
    for (int i = 0; i < score.size(); i++) {
        row++;
        if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
            m_searchEngine->updateSearchCache(score.at(i).pairName,
                                                    SearchEngine::SearchType::DrawnPairsScore);
        }
        color++;
        color = color % 2;
        m_scoreList->setItem(row, 0, new Cell(QString::number(score.at(i).table), color));
        m_scoreList->setSpan(row, 0, 2, 1);
        m_scoreList->setItem(row, 1, new Cell(score.at(i).pairName, color));
        m_scoreList->setItem(row, 2, new Cell(m_db->boogify(score.at(i).score), color));
        for (int booger = 2; booger <= m_db->getBoogersPerRound(); booger++) {
            m_scoreList->setItem(row, 1 + booger, new Cell(m_db->boogify(score.at(++i).score),
                                                           color));
        }

        row++;
        i++;
        if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
            m_searchEngine->updateSearchCache(score.at(i).pairName,
                                                    SearchEngine::SearchType::DrawnPairsScore);
        }
        m_scoreList->setItem(row, 1, new Cell(score.at(i).pairName, color));
        m_scoreList->setItem(row, 2, new Cell(m_db->boogify(score.at(i).score), color));
        for (int booger = 2; booger <= m_db->getBoogersPerRound(); booger++) {
            m_scoreList->setItem(row, 1 + booger, new Cell(m_db->boogify(score.at(++i).score),
                                                           color));
        }
    }
    m_scoreList->resizeColumnsToContents();

    int openRound = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    if (m_roundsState.value(openRound)) {
        m_roundSummaryBox->setTitle(tr("Alle Ergebnisse der %1. Runde").arg(
                                       QString::number(openRound)));
    } else {
        int enteredScores = (row + 1) / 2;
        if (enteredScores == 0) {
            m_roundSummaryBox->setTitle(
                tr("Bisher keine Ergebnisse aus der %1. Runde").arg(
                    QString::number(openRound)));
        } else {
            m_roundSummaryBox->setTitle(
                tr("Bisher %n Ergebnis(se) aus der %1. Runde", "", enteredScores).arg(openRound)
                + QStringLiteral(" ")
                + tr("(%n fehlt/fehlen noch)", "", m_tableSelect->count()));
        }
    }
}

void ScorePage::tableChanged(int tableNumber)
{
    if (! m_autoSelectPairs->isChecked()) {
        return;
    }

    int table = m_tableSelect->itemData(tableNumber).toInt();
    if (m_assignment.assignedPair.contains(table)) {
        if (m_assignment.assignedPair[table].contains(1)) {
            int target = m_playersSelect[1][1]->findData(m_assignment.assignedPair[table][1]);
            if (target != -1) {
                m_playersSelect[1][1]->setCurrentIndex(target);
            }
        }
        if (m_assignment.assignedPair[table].contains(2)) {
            int target = m_playersSelect[2][1]->findData(m_assignment.assignedPair[table][2]);
            if (target != -1) {
                m_playersSelect[2][1]->setCurrentIndex(target);
            }
        }
    }
}

void ScorePage::playersChanged(int pair, int index)
{
    if (! m_autoSelectPairs->isChecked()) {
        return;
    }

    int pairId = m_playersSelect[pair][1]->itemData(index).toInt();
    if (m_assignment.assignedPairNumber.contains(pairId)) {
        if (pair != m_assignment.assignedPairNumber.value(pairId)) {
            QMessageBox::warning(this,
                tr("Die Paarauswahl passt nicht zur Auslosung"),
                tr("<p>Als Paar %1 wurde gerade das Paar „%2“ ausgewählt.</p>"
                    "<p>Laut der Auslosung (bzw. dem Ergebnis der letzten Runde) "
                    "ist dieses Paar aber ein Paar %3!</p>").arg(
                    QString::number(pair),
                    m_playersSelect[pair][1]->currentText(),
                    QString::number(m_assignment.assignedPairNumber.value(pairId))));
        } else {
            if (m_assignment.assignedTable.contains(pairId)) {
                int target = m_tableSelect->findData(m_assignment.assignedTable.value(pairId));
                if (target != -1) {
                    m_tableSelect->setCurrentIndex(target);
                }
            }
        }
    }
}

void ScorePage::disableSaveScore()
{
    m_saveScoreButton->setEnabled(false);
}

void ScorePage::checkSavePossible()
{
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        m_saveScoreButton->setEnabled(m_playersSelect[1][1]->currentIndex()
                                      != m_playersSelect[2][1]->currentIndex());
        break;
    case Database::TournamentMode::SinglePlayers:
        m_saveScoreButton->setEnabled(
            m_playersSelect[1][1]->currentIndex() != m_playersSelect[1][2]->currentIndex()
            && m_playersSelect[1][1]->currentIndex() != m_playersSelect[2][1]->currentIndex()
            && m_playersSelect[1][1]->currentIndex() != m_playersSelect[2][2]->currentIndex()
            && m_playersSelect[1][2]->currentIndex() != m_playersSelect[2][1]->currentIndex()
            && m_playersSelect[1][2]->currentIndex() != m_playersSelect[2][2]->currentIndex()
            && m_playersSelect[2][1]->currentIndex() != m_playersSelect[2][2]->currentIndex()
        );
        break;
    }
}

void ScorePage::saveScore()
{
    emit statusUpdate(tr("Speichere Spielergebnis …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    bool updatePlayersPage = (! m_db->tournamentStarted());

    QList<int> players;
    QList<int> pair1Scores;
    QList<int> pair2Scores;

    for (int i = 1; i <= 2; i++) {
        players << m_playersSelect[i][1]->itemData(m_playersSelect[i][1]->currentIndex()).toInt();
    }
    if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
        for (int i = 1; i <= 2; i++) {
            players << m_playersSelect[i][2]->itemData(
                           m_playersSelect[i][2]->currentIndex()).toInt();
        }
    }

    for (BoogerWidget *boogerWidget : m_boogerWidgets) {
        pair1Scores << boogerWidget->score(1);
        pair2Scores << boogerWidget->score(2);
    }

    int round = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    int table = m_tableSelect->itemData(m_tableSelect->currentIndex()).toInt();
    m_db->saveScore(round, table, players, pair1Scores, pair2Scores);

    QString tableIdentifier;
    if (! m_noTableNumbers->isChecked()) {
        tableIdentifier = tr("Tisch %1").arg(table);
    } else {
        switch (m_db->getTournamentMode()) {
        case Database::TournamentMode::FixedPairs:
            tableIdentifier = formatOpponentsString(
                m_playersSelect[1][1]->itemText(m_playersSelect[1][1]->currentIndex()),
                m_playersSelect[2][1]->itemText(m_playersSelect[2][1]->currentIndex())
            );
            break;
        case Database::TournamentMode::SinglePlayers:
            tableIdentifier = formatOpponentsString(
                QStringLiteral("%1 / %2").arg(
                    m_playersSelect[1][1]->itemText(m_playersSelect[1][1]->currentIndex()),
                    m_playersSelect[1][2]->itemText(m_playersSelect[1][2]->currentIndex())
                ),
                QStringLiteral("%1 / %2").arg(
                    m_playersSelect[2][1]->itemText(m_playersSelect[2][1]->currentIndex()),
                    m_playersSelect[2][2]->itemText(m_playersSelect[2][2]->currentIndex())
                )
            );
            break;
        }
    }

    if (updatePlayersPage) {
        emit tournamentStarted();
    }

    emit scoreChanged();
    reload(round);

    // Make sure the entered score is visible
    int targetRow = 0;
    for (int i = 0; i < m_scoreList->rowCount(); i += 2) {
        if (m_scoreList->item(i, 0)->text().toInt() == table) {
            targetRow = i;
            break;
        }
    }
    if (targetRow == m_scoreList->rowCount() - 2) {
        m_scoreList->scrollToBottom();
    } else {
        m_scoreList->scrollToItem(m_scoreList->item(targetRow, 1),
                                  QAbstractItemView::PositionAtCenter);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Spielergebnis für %1 in Runde %2 gespeichert").arg(
                         tableIdentifier, QString::number(round)));
}

void ScorePage::showContextMenu(const QPoint &point)
{
    if (! m_db->isWritable()) {
        return;
    }

    QTableWidgetItem *selectedScore = m_scoreList->itemAt(point);
    if (selectedScore == nullptr) {
        return;
    }

    int row = selectedScore->row();
    if (row % 2 != 0) {
        row = row - 1;
    }

    QString table = m_scoreList->item(row, 0)->text();
    if (m_noTableNumbers->isChecked()) {
        m_contextMenu->setTitle(formatOpponentsString(m_scoreList->item(row, 1)->text(),
                                                      m_scoreList->item(row + 1, 1)->text()));
    } else {
        m_contextMenu->setTitle(tr("Ergebnis für Tisch %1").arg(table));
    }

    m_selectedTable = table.toInt();

    QPoint targetPoint = point;
    targetPoint.setY(targetPoint.y() + m_scoreList->horizontalHeader()->height());
    m_contextMenu->exec(m_scoreList->mapToGlobal(targetPoint));
}

void ScorePage::deleteScore()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Spielergebnis löschen"));
    messageBox.setText(tr("Soll das Spielergebnis von %1 wirklich gelöscht werden?").arg(
                          getStatusUpdateIdentifier()));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    doDeleteScore(m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt(), m_selectedTable);
}

void ScorePage::doDeleteScore(int round, int table)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche Spielergebnis …"));

    m_db->deleteScore(round, table);
    reload(round);

    if (! m_db->tournamentStarted()) {
        // Re-emitting this will cause the players page to be updated again,
        // as it checks for db.tournamentStarted() itself
        emit tournamentStarted();
    }

    emit scoreChanged();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Spielergebnis für %1 in Runde %2 gelöscht").arg(
                         getStatusUpdateIdentifier(), QString::number(round)));
}

void ScorePage::editScore()
{
    int round = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    QList<Database::SingleScore> score = m_db->getSingleScore(round, m_selectedTable);
    doDeleteScore(round, m_selectedTable);
    m_roundSelect->setCurrentIndex(m_roundSelect->findData(round));

    blockSelectSignals(true);
    m_tableSelect->setCurrentIndex(m_tableSelect->findData(m_selectedTable));
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        setPairOrPlayer(1, 1, score[0].pairOrPlayer);
        setPairOrPlayer(2, 1, score[m_db->getBoogersPerRound()].pairOrPlayer);
        break;
    case Database::TournamentMode::SinglePlayers:
        setPairOrPlayer(1, 1, score[0].pairOrPlayer);
        setPairOrPlayer(1, 2, score[m_db->getBoogersPerRound()].pairOrPlayer);
        setPairOrPlayer(2, 1, score[m_db->getBoogersPerRound() * 2].pairOrPlayer);
        setPairOrPlayer(2, 2, score[m_db->getBoogersPerRound() * 3].pairOrPlayer);
        break;
    }
    blockSelectSignals(false);


    // Get the scores to set
    const QList<int> scores = getBoogerScores(score);

    // Check if we have an aborted score
    int abortedBooger = -1;
    for (int i = 0; i < m_boogerWidgets.count(); i++) {
        if (scores.at(i) < m_db->getBoogerScore()
            && scores.at(i + scores.count() / 2) < m_db->getBoogerScore()) {
            abortedBooger = i;
            break;
        }
    }

    // First reset all boogers
    for (BoogerWidget *booger : m_boogerWidgets) {
        booger->reset();
    }

    // Then mark the first aborted booger if we have one
    if (abortedBooger > -1) {
        boogerAbortedStateChanged(abortedBooger, true);
        m_boogerWidgets.at(abortedBooger)->setAborted(true);
        m_boogerWidgets.at(abortedBooger)->setAsFirstAborted();
    }

    // Pre-set all scores
    for (int i = 0; i < m_boogerWidgets.count(); i++) {
        m_boogerWidgets.at(i)->setScore(scores.at(i), scores.at(i + scores.count() / 2));
    }

    m_saveScoreButton->setEnabled(true);
    m_saveScoreButton->setFocus();
}

QList<int> ScorePage::getBoogerScores(const QList<Database::SingleScore> &score)
{
    QList<int> boogerScores;
    QListIterator<Database::SingleScore> singleScore(score);
    while (singleScore.hasNext()) {
        boogerScores << singleScore.next().score;
        for (int i = 1; i < m_db->getBoogersPerRound(); i++) {
            boogerScores << singleScore.next().score;
        }
        if (m_db->getTournamentMode() == Database::TournamentMode::SinglePlayers) {
            for (int i = 0; i < m_db->getBoogersPerRound(); i++) {
                singleScore.next();
            }
        }
    }
    return boogerScores;
}

void ScorePage::setPairOrPlayer(int pair, int player, int pairOrPlayer)
{
    m_playersSelect[pair][player]->setCurrentIndex(
        m_playersSelect[pair][player]->findData(pairOrPlayer)
    );
}

void ScorePage::blockSelectSignals(bool state)
{
    m_roundSelect->blockSignals(state);
    m_tableSelect->blockSignals(state);
    for (int i = 1; i <= 2; i++) {
        for (int j = 1; j <= 2; j++) {
            m_playersSelect[i][j]->blockSignals(state);
        }
    }
}

void ScorePage::checkEnterScore()
{
    if (! m_saveScoreButton->isEnabled()) {
        return;
    }
    saveScore();
}

void ScorePage::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    setScrollAreaHeight();
}

void ScorePage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_scoreList->rowCount(); row++) {
            if (m_scoreList->isRowHidden(row)) {
                m_scoreList->showRow(row);
            }
        }
        m_scoreList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    m_searchEngine->setSearchTerm(searchText);
    for (int row = 0; row < m_scoreList->rowCount(); row += 2) {
        bool rowMatches = false;
        switch (m_db->getTournamentMode()) {
        case Database::TournamentMode::FixedPairs:
            rowMatches =    m_searchEngine->checkMatch(m_scoreList->item(row,     1)->text())
                         || m_searchEngine->checkMatch(m_scoreList->item(row + 1, 1)->text());
            break;
        case Database::TournamentMode::SinglePlayers:
            rowMatches =    m_searchEngine->checkMatch(m_scoreList->item(row,     1)->text(),
                                SearchEngine::SearchType::DrawnPairsScore)
                         || m_searchEngine->checkMatch(m_scoreList->item(row + 1, 1)->text(),
                                SearchEngine::SearchType::DrawnPairsScore);
            break;
        }
        if (rowMatches) {
            if (m_scoreList->isRowHidden(row)) {
                m_scoreList->showRow(row);
                m_scoreList->showRow(row + 1);
            }
            matches++;
        } else {
            if (! m_scoreList->isRowHidden(row)) {
                m_scoreList->hideRow(row);
                m_scoreList->hideRow(row + 1);
            }
        }
    }

    m_searchWidget->updateMatches(matches);
    m_scoreList->resizeColumnsToContents();

    if (! m_isSearchLabel->isVisible()) {
        m_isSearchLabel->show();
    }
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
}

void ScorePage::swapPairs()
{
    int round = m_roundSelect->itemData(m_roundSelect->currentIndex()).toInt();
    QList<Database::SingleScore> score = m_db->getSingleScore(round, m_selectedTable);
    m_db->deleteScore(round, m_selectedTable);

    QList<int> boogerScores = getBoogerScores(score);
    QList<int> players;
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        players << score[m_db->getBoogersPerRound()].pairOrPlayer
                << score[0].pairOrPlayer;
        break;
    case Database::TournamentMode::SinglePlayers:
        players << score[m_db->getBoogersPerRound() * 2].pairOrPlayer
                << score[0].pairOrPlayer
                << score[m_db->getBoogersPerRound() * 3].pairOrPlayer
                << score[m_db->getBoogersPerRound()].pairOrPlayer;
        break;
    }
    m_db->saveScore(round, m_selectedTable, players,
                    boogerScores.mid(boogerScores.length() / 2, boogerScores.length() / 2),
                    boogerScores.mid(0, boogerScores.length() / 2));

    emit statusUpdate(tr("Paar 1 und Paar 2 für %1 in Runde %2 vertauscht").arg(
                         getStatusUpdateIdentifier(), QString::number(round)));
    reload(round);
}

void ScorePage::noTableNumbersChanged(bool state)
{
    m_scoreList->setColumnHidden(0, state);
    m_tableLabel->setVisible(! state);
    m_tableSelect->setVisible(! state);
    if (state) {
        m_lastAutoSelectState = m_autoSelectPairs->isChecked();
        m_autoSelectPairs->setChecked(false);
        m_autoSelectPairs->hide();
    } else {
        m_autoSelectPairs->show();
        m_autoSelectPairs->setChecked(m_lastAutoSelectState);
    }

    m_db->setNoTableNumbers(state);
}

QString ScorePage::getStatusUpdateIdentifier()
{
    return m_noTableNumbers->isChecked() ? m_contextMenu->title()
                                         : tr("Tisch %1").arg(QString::number(m_selectedTable));
}

void ScorePage::autoSelectPairsChanged(bool state)
{
    m_db->setAutoSelectPairs(state);
    if (state) {
        tableChanged(m_tableSelect->currentIndex());
    }
}

QString ScorePage::formatOpponentsString(QString pair1, QString pair2)
{
    if (pair1.size() > 15) {
        pair1 = pair1.left(15) + QStringLiteral("…");
    }
    if (pair2.size() > 15) {
        pair2 = pair2.left(15) + QStringLiteral("…");
    }
    return tr("„%1“ gegen „%2“").arg(pair1, pair2);
}

void ScorePage::setNoTableNumbers(bool state)
{
    m_noTableNumbers->setChecked(state);
}

void ScorePage::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs->setChecked(state);
}

void ScorePage::checkCorrectPairSelection()
{
    if (! m_resultBox->isVisible()) {
        return;
    }

    if (m_autoSelectPairs->isChecked()) {
        tableChanged(0);
    }
}

void ScorePage::boogerAbortedStateChanged(int booger, bool aborted)
{
    if (aborted) {
        for (int i = booger + 1; i < m_boogerWidgets.count(); i++) {
            m_boogerWidgets.at(i)->setAborted(true);
        }
    } else {
        for (int i = 0; i <= booger; i++) {
            m_boogerWidgets.at(i)->reset();
        }
        if (booger + 1 < m_boogerWidgets.count()) {
            m_boogerWidgets.at(booger + 1)->setAsFirstAborted();
        }
    }
}

void ScorePage::resetSearch()
{
    m_searchWidget->clear();
}

void ScorePage::showScoreDialog(BoogerSpinBox *spinBox)
{
    m_targetSpinBox = spinBox;
    m_scoreDialog->show();
}

void ScorePage::setSelectedScore()
{
    m_targetSpinBox->setValue(m_scoreDialog->score());
    m_targetSpinBox->setFocus();
    m_targetSpinBox->selectAll();
}
