/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "SearchEngine.h"
#include "StringTools.h"

// Qt includes
#include <QRegularExpression>
#include <QDebug>

namespace {

const QList<QChar> CHARS_TO_REPLACE({
    QChar(0x00e4), // ä
    QChar(0x00f6), // ö
    QChar(0x00fc), // ü
    QChar(0x00df), // ß
    QChar(0x00E6), // æ
    QChar(0x0153), // œ
    QChar(0x00F8)  // ø
});

const QStringList CHAR_REPLACEMENTS({
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("ue"),
    QStringLiteral("ss"),
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("oe")
});

}

SearchEngine::SearchEngine(QObject *parent, StringTools *stringTools)
    : QObject(parent), m_stringTools(stringTools)
{
}

const QStringList SearchEngine::searchVariants(const QString &text) const
{
    // Create a sorted version of the search string split by words
    QStringList words = m_stringTools->toLower(text).split(QStringLiteral(" "));
    m_stringTools->sortList(words);
    const QString sorted = words.join(QStringLiteral(" "));

    // Generate all variants
    QStringList variants;
    m_stringTools->appendIfNotExists(variants, sorted);
    const QString diacriticsStripped = stripDiacritis(sorted);
    m_stringTools->appendIfNotExists(variants, diacriticsStripped);
    m_stringTools->appendIfNotExists(variants, replaceUmlauts(diacriticsStripped));
    const QString lowerReplaced = replaceUmlauts(sorted);
    m_stringTools->appendIfNotExists(variants, lowerReplaced);
    m_stringTools->appendIfNotExists(variants, stripDiacritis(lowerReplaced));

    return variants;
}

QString SearchEngine::replaceUmlauts(QString text) const
{
    for (int i = 0; i < CHARS_TO_REPLACE.size(); i++) {
        text.replace(CHARS_TO_REPLACE.at(i), CHAR_REPLACEMENTS.at(i));
    }
    return text;
}

QString SearchEngine::stripDiacritis(const QString &text) const
{
    // Here, we strip out all diacritical marks and accents by first separating the combinig
    // characters and then removing them, cf. http://en.wikipedia.org/wiki/Combining_character
    QString replaced = text.normalized(QString::NormalizationForm_D);
    for (int i = 0x0300; i <= 0x036F; i++) {
        replaced.remove(QChar(i));
    }
    return replaced;
}

bool SearchEngine::checkMatch(const QString &text, SearchType searchType) const
{
    switch (searchType) {
    case SearchType::Default:
        return checkMatch(text, m_playersNamesVariants);
    case SearchType::DrawnPairsScore:
        return checkMatch(text, m_drawnPairsNamesVariants);
    }

    Q_UNREACHABLE();
    return false;
}

bool SearchEngine::checkMatch(const QString &text,
                              const QMap<QString, QStringList> &variantsList) const
{
    QRegularExpression re;
    for (const QString &searchTermRegularExpression : m_searchTermRegularExpressions) {
        re.setPattern(searchTermRegularExpression);
        for (const QString &targetVariant : variantsList.value(text)) {
            if (re.match(targetVariant).hasMatch()) {
                return true;
            }
        }
    }
    return false;
}

void SearchEngine::clearSearchCache(SearchType searchType)
{
    switch (searchType) {
    case Default:
        m_playersNamesVariants.clear();
        break;
    case DrawnPairsScore:
        m_drawnPairsNamesVariants.clear();
        break;
    }
}

void SearchEngine::updateSearchCache(const QString &name, SearchType searchType)
{
    switch (searchType) {
    case Default:
        m_playersNamesVariants.insert(name, searchVariants(name));
        break;
    case DrawnPairsScore:
        m_drawnPairsNamesVariants.insert(name, searchVariants(name));
        break;
    }
}

void SearchEngine::setSearchTerm(const QString &text)
{
    m_searchTermRegularExpressions.clear();
     for (const QString &variant : searchVariants(text.simplified())) {
        QStringList words = variant.split(QStringLiteral(" "));
        for (QString &word : words) {
            word = QRegularExpression::escape(word);
        }
        m_searchTermRegularExpressions.append(words.join(QStringLiteral(".+")));
     }
}

QMap<QString, QString> SearchEngine::checkListForPossibleDuplicates(const QStringList &list)
{
    QMap<QString, QString> possibleDuplicates;
    QMap<QString, QStringList> variantsCache;
    for (QString name : list) {
        variantsCache.insert(name, searchVariants(name.simplified()));
    }

    for (const QString &searchName : list) {
        setSearchTerm(searchName);
        for (const QString &listName : list) {
            if (searchName == listName) {
                continue;
            }
            if (checkMatch(listName.simplified(), variantsCache)) {
                possibleDuplicates[searchName] = listName;
                break;
            }
        }
    }
    return possibleDuplicates;
}
